/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <wfs_misc.h>
#include "wfs_cacheloc.h"

#include <assert.h>

wfs_cacheloc::wfs_cacheloc() {

    assert( pthread_mutex_init( &_writes_lock, NULL ) == 0 );
}

wfs_cacheloc::~wfs_cacheloc() {
    pthread_mutex_destroy( &_writes_lock );
}

bool wfs_cacheloc::get_close_nodes( wfs_file_id fid, wfs_version v,
                                    vivaldi *viv, const Coord &which,
                                    unsigned long long offset, int n,
                                    std::vector<cacheloc_ent> &list,
				    unsigned sz) {

    uint64_t key = (uint64_t) fid;

    //TODO: figure out how to incorporate node id into key
    key = (key << 32) | v.version_num;

    //bool r = false;

    pthread_mutex_lock( &_writes_lock );
    std::vector<cacheloc> *vv = _locs[key];

    if (vv) {
        for (unsigned i = 0; i < vv->size(); i++) {
            for (unsigned k = 0; k < ((*vv)[i]).off.size(); k++) {

                unsigned long long curroff = ((*vv)[i]).off[k];
                unsigned long long currn = ((*vv)[i]).n[k];

                if (curroff <= offset &&
                    (curroff + currn) >= (offset+n)) {
                    std::vector<cacheloc_ent>::iterator j;
                    double dist;
                    if (viv)
                        dist = viv->get_distance_to_ip((*vv)[i].s_addr, (*vv)[i].s_port);
                    else
                        dist = (*vv)[i].c.distance_f(which);
                    for (j = list.begin(); j!= list.end(); j++) {
                        if ((viv &&
                             viv->get_distance_to_ip((*j).s_addr, (*j).s_port) > dist)
                            || (!viv && (*j).c.distance_f(which) > dist)) {
                            break;
                        }
                    }
                    /**
                       in_addr sin_addr;
                       sin_addr.s_addr = (*vv)[i].s_addr;
                       printf( "get close nodes on %016"PRIx64" considering, addr %s:%d ",
                       fid, inet_ntoa(sin_addr), ntohs((*vv)[i].s_port) );
                       for( uint k1 = 0; k1 < ((*vv)[i]).off.size(); k1++ ) {
                       printf( "(%llu,%llu) ", ((*vv)[i]).off[k1], ((*vv)[i]).n[k1] );
                       }
                       printf( " distance %.3f\n", dist );
                    */
                    list.insert(j, (*vv)[i]);
                    if (list.size() > sz)
                        list.pop_back();
                }
            }
        }
    }
    double now = vivaldi::gettime();
    for (unsigned i = 0; i < list.size(); i++)
        list[i].age = (uint32_t)(now - list[i].when)+ list[i].age;
    pthread_mutex_unlock( &_writes_lock );
    return (list.size() == sz);
}

bool wfs_cacheloc::get_rand_nodes(wfs_file_id fid, wfs_version v,
                                  unsigned long long offset, int n,
                                  std::vector<cacheloc_ent> &list,
				  unsigned sz) {

    //TODO: figure out how to incorporate node id into key
    uint64_t key = (uint64_t) fid;
    key = (key << 32) | v.version_num;

    //bool r = false;

    pthread_mutex_lock( &_writes_lock );
    std::vector<cacheloc> *vv = _locs[key];

    if (vv) {
        int s = random() % vv->size();
        for (unsigned i = 0; i < vv->size(); i++) {
            for (unsigned k = 0; k < ((*vv)[s]).off.size(); k++) {

                unsigned long long curroff = ((*vv)[s]).off[k];
                unsigned long long currn = ((*vv)[s]).n[k];

                if (curroff <= offset &&
                    (curroff + currn) >= (offset+n)) {
                    list.push_back((*vv)[s]);
                    if (list.size() == sz) {
                        i = vv->size();
                        break;
                    }
                }
                s = (s+1) % vv->size();
            }
        }
    }
    double now = vivaldi::gettime();
    for (unsigned i = 0; i < list.size(); i++)
        list[i].age = (uint32_t)(now - list[i].when)+ list[i].age;

    pthread_mutex_unlock( &_writes_lock );
    /*
      if (vv && vv->size() > 0) {
      printf("get_rand_nodes vv[0].off %llu vv[0].n %llu off %llu n %d list.size %d requested %d\n",
      (*vv)[0].off, (*vv)[0].n, offset, n, list.size(), sz);
      }
    */
    return (list.size() == sz);
}

#ifndef max
#define max(a,b) (a>b)?a:b
#endif

void wfs_cacheloc::update_cacheloc(wfs_file_id fid, wfs_version v,
                                   std::vector<cacheloc_ent> const &list)
{
    if (list.size() == 0 || list[0].n[0] == 0)
        return;
    uint64_t key = (uint64_t) fid;
    //TODO: figure out how to incorporate node id into key
    key = (key << 32) | v.version_num;

    pthread_mutex_lock( &_writes_lock );
    std::vector<cacheloc> *vv = _locs[key];
    if (!vv) {
        vv = new std::vector<cacheloc>;
        _locs[key] = vv;
    }

    double now = vivaldi::gettime();
    for (unsigned i = 0; i < list.size(); i++) {
      unsigned j = 0;
      for (j = 0; j < vv->size(); j++) {
        if ((*vv)[j].s_addr == list[i].s_addr
            && (*vv)[j].s_port == list[i].s_port) {
          (*vv)[j].c = list[i].c;
          (*vv)[j].when = now;
          (*vv)[j].age = list[i].age;
#ifdef WFS_WITH_LIBSSH
          std::map<unsigned,cacheloc::checksum>::const_iterator it;
          for (it = list[i].checksums.begin();
          it != list[i].checksums.end(); it++) {
            (*vv)[j].checksums[it->first] = it->second;
          }
#endif
          // now see if we can patch up any of the byte ranges here
          for( uint k2 = 0; k2 < list[i].off.size(); k2++ ) {

            bool added = false;
            unsigned long long listoff = list[i].off[k2];
            unsigned long long listn = list[i].n[k2];

            for( uint k1 = 0; k1 < ((*vv)[j]).off.size(); k1++ ) {

              unsigned long long curroff = ((*vv)[j]).off[k1];
              unsigned long long currn = ((*vv)[j]).n[k1];

              if( listoff == curroff && listn == currn ) {
                added = true;  // duplicate
                break;
              } else if (curroff < listoff) {
                if ((curroff+currn) >= listoff) {
                  (*vv)[j].n[k1] = max(listoff + listn - curroff, currn);
                  added = true;
                  break;
                }
              }else{
                if ((listoff+listn) >= curroff) {
                  (*vv)[j].off[k1] = listoff;
                  (*vv)[j].n[k1] = max(curroff + currn - listoff, listn);
                  added = true;
                  break;
                }
              }
            }

            if( !added ) {
              (*vv)[j].off.push_back(listoff);
              (*vv)[j].n.push_back(listn);
            }

          }
          // TODO: coalesce holes if needed -- only an optimization
          // as long as reads always occur along uniform block boundaries
          if( wfs_dbg_block(WFS_DBG_3) ) {
            in_addr sin_addr;
            sin_addr.s_addr = (*vv)[j].s_addr;
            wfs_log(WFS_DBG_3, "update_cache_loc called on %016"PRIx64", addr %s:%d",
                fid, inet_ntoa(sin_addr), ntohs((*vv)[j].s_port));
            for( uint k1 = 0; k1 < ((*vv)[j]).off.size(); k1++ ) {
              wfs_log(WFS_DBG_3, "\t\t\t\(%llu,%llu) ", ((*vv)[j]).off[k1], ((*vv)[j]).n[k1] );
            }
          }
          break;
        }
      }
      if (j  == vv->size()) {
        vv->push_back(list[i]);
        if(wfs_dbg_block(WFS_DBG_3) ) {
          in_addr sin_addr;
          sin_addr.s_addr = list[i].s_addr;
          wfs_log(WFS_DBG_3, "update_cache_loc (new) called on %016"PRIx64", addr %s:%d",
              fid, inet_ntoa(sin_addr), ntohs(list[i].s_port));
          for( uint k1 = 0; k1 < list[i].off.size(); k1++ ) {
            wfs_log(WFS_DBG_3, "\t\t\t(%llu,%llu) ", list[i].off[k1], list[i].n[k1] );
          }
        }
      }
      assert(vv->size() > 0);
    }
    pthread_mutex_unlock( &_writes_lock );
}

