/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


//
// Replicated state machine implementation with a master and several
// slaves. The master receives requests, assigns them a view stamp (a
// vid, and a sequence number) in the order of reception, and forwards
// them to all slaves. A slaves executes requests in the order of that
// the master stamps them and replies with an OK to the master. The
// master executes the request after it receives OKs from all slaves,
// and sends the reply back to the client.
//
// After a failure, and during recovery, each member of the RSM
// contacts all other members (provided by Paxos) to find who executed
// the most recent operation (based on the viewstamp).  The most
// recent one will be the new master.  Ties are broken by chosing the
// lowest numbered member.
//
// The implementation should be used only with servers that run all
// requests run to completion; in particular, a request shouldn't
// block.  If a request blocks, the slaves won't respond to the
// master, and the master won't execute the request.  A request may
// send an RPC to another host, but the RPC should be a one-way
// message to that host; the slaves shouldn't do anything based on the
// response or execute after the response, because it is not
// guaranteed that all slaves will receive the same response and
// execute in the same order.
//
//

#include "rsm.h"
#include "rsm_client.h"
#include <fstream>
#include <iostream>
#include <method_thread.h>
#include <signal.h>
#include <sys/time.h>
#include <time.h>
#include "wfs_misc.h"


static void *
recoverythread(void *x)
{
  rsm *r = (rsm *) x;
  r->recovery();
  return 0;
}

static bool break1 = false;

static void
breakpoint1()
{
  if (break1) {

  	wfs_log(WFS_DBG_1,"Dying at breakpoint 1 in rsm!\n");

    exit(1);
  }
}

// This works only correctly when there's one RSM object per process
void handle_int(int param)
{

  if (param == SIGINT) {

  	wfs_log(WFS_DBG_1,"Got SIGINT\n");

    break1 = true;
  }
}

rsm::rsm(std::string _first, std::string _me, std::string log_dir,
	 wfs_client_manager *clients, int pxs_delay_max)
  : stf(0), master(_first), stable (false), ops_since_last_cp(0),
    invoke_rec_head(NULL), invoke_rec_tail(NULL)
{

  pthread_mutex_init(&rsm_mutex, NULL);
  pthread_cond_init(&recovery_cond, NULL);

  pthread_mutex_init(&invoke_rec_mutex, NULL);

  pxs = new paxos(_first, _me, log_dir, clients, pxs_delay_max);
  rlog = new rsm_log(_me, log_dir);

  //signal( SIGINT, handle_int);

  rsmrpc = pxs->get_rpcs();
  rsmrpc->reg(rsm_client_protocol::invoke, this, &rsm::client_invoke);
  rsmrpc->reg(rsm_client_protocol::members, this, &rsm::client_members);
  rsmrpc->reg(rsm_protocol::invoke, this, &rsm::invoke);
  rsmrpc->reg(rsm_protocol::cptransferreq, this, &rsm::cptransferreq);
  rsmrpc->reg(rsm_protocol::logtransferreq, this, &rsm::logtransferreq);
  rsmrpc->reg(rsm_protocol::viewstampreq, this, &rsm::viewstampreq);

}

rsm::~rsm()
{
	delete pxs;
	delete rlog;
}

rsm_handler::rsm_handler()
	: _usedecision(false)
{
}


void
rsm::reg1(int proc, rsm_handler *h)
{
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  procs[proc] = h;
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
}

//called when a node first starts up, it reads from disk
//to re-execute saved log requests
void
rsm::recover_from_log()
{
	pthread_mutex_lock(&rsm_mutex);
	pthread_t th;
	viewstamp vs(pxs->vid(),0);
	rsm_record rec;
	if (rlog->load_checkpoint(rec) && stf) {
		stf->unmarshal_state(rec.req);
		last_myvs = myvs = rec.vs;
	}

	wfs_log(WFS_DBG_3,"rsm::recover_from_log recover from cp %d,%d\n", last_myvs.vid, last_myvs.seqno);

  	int n = 0;
	while (rlog->load_request(vs,rec)) {
		vs = rec.vs;
		n++;
		execute(rec.procno,rec.req);
	}
	last_myvs = myvs = vs;
	stable = true;

	wfs_log(WFS_DBG_3,"rsm::recover_from_log recover from %d log entries myvs %d,%d \n", n, myvs.vid, myvs.seqno);


	pxs->set_paxos_view_change(this);
	//recovery thread is in charge of performing state transfer from new master
	assert(pthread_create(&th, NULL, &recoverythread, (void *) this) == 0);
	pthread_mutex_unlock(&rsm_mutex);
}

void
rsm::recovery()
{
  struct timespec next_timeout;
  struct timeval tp;

  assert(pthread_mutex_lock(&rsm_mutex)==0);
  while(1) {
    //This node needs to perform state transfer
    //if paxos has finished performing a view change.
    //For state transfer, this node must first find the master
    //which is the node with the largest viewstamp number (with ties broken by id)
    //and then transfer state from the master.
    if (!stable){

      wfs_log(WFS_DBG_3,"recovery: stable %d pxs stable %d vs (%d,%d) vid %d\n", stable?1:0,
          pxs->is_stable(), myvs.vid, myvs.seqno, pxs->vid());

    }
    if (pxs->is_stable() && !stable) {
      viewstamp vs;
      if (find_master(vs)) {
        if (vs > myvs) {
          //if (statetransfer(master, vs)) {
					if (logtransfer(master,vs)) {
            myvs = vs;
            myvs.vid = pxs->vid();
            myvs.seqno = 0;
            stable = true;

            wfs_log(WFS_DBG_3,"recovery: stable master %s vs (%d,%d)\n",
                	master.c_str(), myvs.vid, myvs.seqno);


          }
        } else {

          wfs_log(WFS_DBG_3,"recovery: stable master %s vs (%d,%d)\n",
              master.c_str(), myvs.vid, myvs.seqno);

          myvs.vid = pxs->vid();
          myvs.seqno = 0;
          stable = true;
        }
      }
    }

    //check if it is time to checkpoint state to disk
    if (rlog->do_checkpoint()) {
      rsm_record r;
      r.procno = 0;
      r.vs = last_myvs;
      r.req = stf->marshal_state();
      rlog->save_checkpoint(r);
    }
    gettimeofday(&tp, NULL);
    next_timeout.tv_sec  = tp.tv_sec;
    next_timeout.tv_nsec = tp.tv_usec * 1000;
    next_timeout.tv_sec += 1;

    wfs_log(WFS_DBG_3,"recovery sleeping\n" );

    pthread_cond_timedwait(&recovery_cond, &rsm_mutex, &next_timeout);

    wfs_log(WFS_DBG_3, "recovery woken\n" );

  }
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
}

bool
rsm::find_master(viewstamp &vs)
{
  std::vector<std::string> mems;
  viewstamp r;
  int ret;

  std::string old_master = master;
  pxs->get_members(mems);
  master = pxs->myaddr();
  vs = last_myvs;
  for (unsigned i = 0; i < mems.size(); i++) {
    if (mems[i] != pxs->myaddr()) {

      wfs_log(WFS_DBG_3,"find_master at %s\n", mems[i].c_str());

      assert (pthread_mutex_unlock(&rsm_mutex) == 0);
      rpcc *cl = pxs->get_cl(mems[i]);
      if( cl == NULL ) {
	ret = rpc_const::timeout_failure;
      } else {
	ret = cl->call(rsm_protocol::viewstampreq, pxs->vid(),
		       r, rpcc::to(10000));
	pxs->return_cl(mems[i]);
      }
      assert (pthread_mutex_lock(&rsm_mutex) == 0);
      if (ret != rsm_protocol::OK) {
        if (ret != rsm_protocol::BUSY) {

          wfs_log(WFS_DBG_2,"find_master: member %s didn't respond %d start paxos\n",
              mems[i].c_str(), ret);

          pxs->start_paxos();
        }
        return false;
      } else {
        if (r > vs) {
          vs = r;

          wfs_log(WFS_DBG_3,"find_master: %s(%d,%d) is master\n", mems[i].c_str(),
              vs.vid, vs.seqno);

          master = mems[i];
        } else if (r == vs) {
          if (mems[i] < master) {

          	wfs_log(WFS_DBG_3,"find_master: break tie; %s(%d,%d) is master\n",
                mems[i].c_str(), vs.vid, vs.seqno);

            master = mems[i];
          }
        }
      }
    }
  }

  // NOTE: rsm_mutex is held over these executions, they are hopefully local
  if( master == pxs->myaddr() && old_master != pxs->myaddr() ) {
    marshall args;
    args << old_master;
    execute( rsm_reserved::became_master, args.str(), false );
  } else if( old_master == pxs->myaddr() && master != pxs->myaddr() ) {
    marshall args;
    args << master;
    execute( rsm_reserved::lost_master, args.str(), false );
  }

  return true;
}

//assumes that rsm_mutext is already held
bool
rsm::logtransfer(std::string m, viewstamp vs)
{
	std::vector<rsm_record> vr;
	rsm_record r;
	int ret;
	rpcc *which_cl = pxs->get_cl(m);
	assert(pthread_mutex_unlock(&rsm_mutex)==0);
	// TODO: we can probably get rid of this call altogether, since
	// logtransferreq takes care of this now
	if( which_cl == NULL ) {
	  ret = rpc_const::timeout_failure;
	} else {
	  ret = which_cl->call(rsm_protocol::cptransferreq, pxs->myaddr(),
			       last_myvs, r, rpcc::to(10000));
	}
    if (ret == rsm_protocol::OK) {

    	wfs_log(WFS_DBG_3,"rsm::logtransfer get checkpoints state from %s %u,%d->%u,%d state sz %u\n",
		       m.c_str(), (unsigned)myvs.vid, myvs.seqno,
		       (unsigned)r.vs.vid, r.vs.seqno,
		       (unsigned)r.req.length());

    	rlog->save_checkpoint(r);
    	assert(stf);
    	stf->unmarshal_state(r.req);
    	last_myvs = myvs = r.vs;
    }else{

    	wfs_log(WFS_DBG_3,"rsm::logtransfer (ret %d) my cp state is at least as new %d,%d %d,%d\n",
		       ret,last_myvs.vid,last_myvs.seqno,r.vs.vid,r.vs.seqno);

    }
    assert(pthread_mutex_lock(&rsm_mutex)==0);

	int transferred = 0;
	do {
		assert(pthread_mutex_unlock(&rsm_mutex)==0);
		if( which_cl == NULL ) {
		  ret = rpc_const::timeout_failure;
		} else {
		  ret = which_cl->call(rsm_protocol::logtransferreq, pxs->myaddr(),
				       last_myvs, vs, vr, rpcc::to(10000));
		  pxs->return_cl(m);
		}
		assert(pthread_mutex_lock(&rsm_mutex)==0);

		wfs_log(WFS_DBG_3,"rsm::logtransfer get from %d,%d - %d,%d from %s ret %d missing sz %u\n", last_myvs.vid, last_myvs.seqno,
				vs.vid,vs.seqno, m.c_str(), ret,
			       (unsigned)vr.size());

		if (ret != rsm_protocol::OK) {
			return false;
		}
		for (unsigned i = 0; i < vr.size(); i++) {
		  if( vr[i].cp ) {
		    rlog->save_checkpoint(vr[i]);
		    assert(stf);
		    stf->unmarshal_state(vr[i].req);
		    last_myvs = myvs = vr[i].vs;
		  } else {
		    rlog->save_request(vr[i], (i == (vr.size()-1)));
		    execute(vr[i].procno, vr[i].req);
		    last_myvs = myvs = vr[i].vs;
		    transferred++;
		  }
		}
		// if there were no records, obviously we can just
		// catch right up
		if( vr.size() == 0 ) {
		  last_myvs = myvs = vs;
		}
	} while (myvs!=vs);


	wfs_log(WFS_DBG_3,"rsm::logtransfer transferred %d records from %s vs now %d,%d\n", transferred,
			m.c_str(), vs.vid, vs.seqno);

	if (myvs == vs)
		return true;
	else
		return false;
}

/**
 * Call to transfer state from m to the local node.
 * Assumes that rsm_mutex is already held.
 */
bool
rsm::statetransfer(std::string m, viewstamp vs)
{
  std::string r;
  rpcc *which_cl = pxs->get_cl(m);
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  int ret;
  if( which_cl == NULL ) {
    ret = rpc_const::timeout_failure;
  } else {
    ret = which_cl->call(rsm_protocol::transferreq, pxs->myaddr(),
			 vs, r, rpcc::to(10000));
    pxs->return_cl(m);
  }
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  if (ret != rsm_protocol::OK)
    return false;
  if (stf) {
    stf->unmarshal_state(r);
  }

  wfs_log(WFS_DBG_3,"rsm::statetransfer transfer from %s success, vs(%d,%d)\n",
   	  m.c_str(), vs.vid, vs.seqno);

  return true;
}

/*
 * Paxos informs rsm whenever it has successfully
 * completed a view change
 */
void
rsm::view_change()
{
  pthread_mutex_lock(&rsm_mutex);

  wfs_log(WFS_DBG_3,"view change\n");

  if (pxs->vid() >= last_myvs.vid) {
  	stable = false;
  	pthread_cond_signal(&recovery_cond);
  }
  pthread_mutex_unlock(&rsm_mutex);
}

void
rsm::insert(arg_t *a)
{
  std::list<arg_t*>::iterator it;
  for (it = requests.begin(); it != requests.end(); it++) {
    if (a->seqno == (*it)->seqno)
      return;
    if (a->seqno < (*it)->seqno) {
      requests.insert(it,a);
      return;
    }
  }
  requests.push_back(a);
}

std::string
rsm::execute(int procno, std::string req, bool replicated)
{
  if( procs.find(procno) == procs.end() ) {
    return "";
  }

  wfs_log(WFS_DBG_3,"rsm::execute %x reqsz %u\n", procno, (unsigned)req.length());

  if(replicated)
    ops_since_last_cp++;
  rsm_handler *h = procs[procno];
  assert(h);
  unmarshall args(req);
  marshall rep;
  rsm_protocol::status ret = h->fn(args, rep);
  marshall rep1;
  rep1 << ret;
  rep1 << rep.str();
  return rep1.str();
}

void
rsm::call_invoke(std::string mem, std::vector<rsm_record> recs)
{
  int ret1;
  rpcc *cl = pxs->get_cl(mem);
  //a backup node might not have finished statetransfer
  //before master start sending requests to do
  //be patient
  int maxsleep = 4;
  int r1;
  while (maxsleep>0) {
    if( cl == NULL ) {
      ret1 = rpc_const::timeout_failure;
    } else {
      ret1 = cl->call(rsm_protocol::invoke, recs,
		      r1, rpcc::to(10000));
    }
    if (ret1!=rsm_protocol::BUSY)
      break;
    sleep(1);
    maxsleep--;
  }
  if( cl != NULL ) {
    pxs->return_cl(mem);
  }

  wfs_log(WFS_DBG_3, "invoke: member %s complete!\n", mem.c_str() );

  pthread_exit( (void *) ret1 );
}

//
// Clients call client_invoke to invoke a procedure on the replicated state
// machine: the master receives the request, assigns it a sequence
// number, and invokes it on all members of the replicated state
// machine.
//
rsm_client_protocol::status
rsm::client_invoke(int procno, std::string req, std::string &r)
{
  int ret = rsm_protocol::OK;
  int tag = random();

  wfs_log(WFS_DBG_3,"rsm::client_invoke %d 0x%x vs (%d,%d) vid %d\n", tag, procno, myvs.vid, myvs.seqno,
      pxs->vid());

  // put this request on the queue.  don't have to worry about
  // the queue getting too big, since the rpc library limits the
  // number of outstanding requests anyway
  invoke_rec irec(procno, req);
  assert(pthread_mutex_lock(&invoke_rec_mutex)==0);
  if( invoke_rec_head == NULL ) {
    invoke_rec_head = &irec;
  }
  if( invoke_rec_tail != NULL ) {
    invoke_rec_tail->next = &irec;
  }
  invoke_rec_tail = &irec;
  assert(pthread_mutex_unlock(&invoke_rec_mutex)==0);

  //now try to be the one who processes the queue
  assert(pthread_mutex_lock(&rsm_mutex)==0);

  invoke_rec *firstrec;
  invoke_rec *ihead;

  // if our record was done already, we're done
  if( irec.ret != invoke_rec::NOTDONE ) {
    goto exit;
  }

  wfs_log(WFS_DBG_3,"rsm::client_invoke starting %d\n", tag);

  // otherwise we're in charge of executing for everyone
  assert(pthread_mutex_lock(&invoke_rec_mutex)==0);
  firstrec = invoke_rec_head;
  ihead = firstrec;
  invoke_rec_head = NULL;
  invoke_rec_tail = NULL;
  assert(pthread_mutex_unlock(&invoke_rec_mutex)==0);

  if (get_master() == pxs->myaddr()) {
    int ret1;

    std::vector<rsm_record> rsm_recs;

    int count = 0;
    while( ihead != NULL ) {

      // Call decision function.
      rsm_handler *h = procs[ihead->procno];
      assert(h);
      if(h->getusedecision()){
        unmarshall args(ihead->req);
        marshall rep,rep1;
        int wccret;
        dec_ret cont = h->decide(args,rep,wccret);
        switch (cont) {
        case NO:
          rep1 << wccret;
          rep1 << rep.str();
          ihead->r = rep1.str();
          ihead->ret = rsm_protocol::OK;
          ihead = ihead->next;
          continue;
        case YES_NOREP:
          // Just Execute and continue;
          ihead->r = execute(ihead->procno, ihead->req,false);
          ihead->ret = rsm_protocol::OK;
          ihead = ihead->next;
          continue;
        case YES_REP:
          break;
        default:
          wfs_log(WFS_DBG_2,"invoke: Decision Function returned unknown code:%d\n",cont);
          break;
        }
      }

      viewstamp s = myvs;
      rsm_record rec(s,ihead->procno,ihead->req);
      last_myvs = myvs;
      myvs.seqno++;
      rsm_recs.push_back(rec);

      //execute at myself
      // (this disk sync will not be in parallel with the other nodes'
      //  execution.  TODO: add an explicit sync method to rsm_log to fix
      //  this.)
      rlog->save_request(rec, (ihead->next == NULL));
      ihead->r = execute(ihead->procno, ihead->req);

      ihead = ihead->next;
      count++;

    }

    if( count == 0 ) {
      goto exit;
    }

    if( count > 1 ) {
      wfs_log(WFS_DBG_3, "client_invoke: Batching together %d invokes\n", count );
    }

    std::vector<std::string> mems;
    pxs->get_members(mems);

    std::map<std::string,pthread_t> pths;
    for (unsigned i = 0; i < mems.size(); i++) {
      if (mems[i] != pxs->myaddr()) {

        wfs_log(WFS_DBG_3,"invoke at %s\n", mems[i].c_str());

        pthread_t pth = method_thread( this, false, &rsm::call_invoke,
            mems[i], rsm_recs );
        assert(pth != 0);
        pths[mems[i]] = pth;
      }
    }

    ret1 = rsm_protocol::OK;
    for (std::map<std::string,pthread_t>::iterator i = pths.begin();
    i != pths.end(); i++) {
      std::string mem = i->first;
      pthread_t pth = i->second;
      void *tmptr;
      int ret2;
      pthread_join(pth, &tmptr);
      ret2 = (int)(long)tmptr;

      if (ret2 != rsm_protocol::OK) {

        wfs_log(WFS_DBG_2,"invoke: member %s is trouble, start paxos\n", mem.c_str());

        ret1 = ret2;
      }
    }

    if( ret1 != rsm_protocol::OK ) {
      pxs->start_paxos();
      ihead = firstrec;
      ret = rsm_client_protocol::BUSY;
      goto set_all_rets;
    }

    breakpoint1();

  } else {
    if (get_master() == "") {
      wfs_log(WFS_DBG_2,"client_invoke at a member who is busy with recovery: pxs stable? %d rsm stable? %d master %s\n",
          pxs->is_stable(), stable, master.c_str());
      ret = rsm_client_protocol::BUSY;
    } else {

      wfs_log(WFS_DBG_2,"client_invoke at a member who is not the master\n");

      ret = rsm_client_protocol::NOTMASTER;
    }
  }

  set_all_rets:
  ihead = firstrec;
  while( ihead != NULL ) {
    if( ihead->ret == invoke_rec::NOTDONE ) {
      ihead->ret = ret;
    }
    ihead = ihead->next;
  }

  exit:
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  wfs_log(WFS_DBG_3,"rsm::client_invoke done %d 0x%x vs (%d,%d) vid %d ret %d\n", tag, procno, myvs.vid, myvs.seqno,
      pxs->vid(), irec.ret);

  r = irec.r;
  return irec.ret;
}

//
// The master calls the internal invoke at each member of the
// replicated state machine
//
// the replica must execute requests in order (with no gaps)
// according to requests' seqno

rsm_protocol::status
rsm::invoke(std::vector<rsm_record> recs, int &dummy)
{
  rsm_protocol::status ret = rsm_protocol::OK;
  std::string rep;
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  if (stable && (get_master() != pxs->myaddr())) {

    for( unsigned i = 0; i < recs.size(); i++ ) {
      rsm_record rec = recs[i];

      wfs_log(WFS_DBG_3,"doinvoke: handler 0x%x vs (%d,%d) myvs (%d,%d)\n",
          rec.procno, rec.vs.vid, rec.vs.seqno, myvs.vid, myvs.seqno);

      if (rec.vs < myvs) {  // old request

        wfs_log(WFS_DBG_3,"doinvoke: old request\n");

        continue;
      } else {
        if (rec.vs == myvs) {
          rlog->save_request(rec, (i == (recs.size()-1)));
          rep = execute(rec.procno, rec.req);
          last_myvs = myvs;
          myvs.seqno++;
        }else{

          wfs_log(WFS_DBG_3,"doinvoke: handler 0x%x request into the future?\n", rec.procno);

          ret = rsm_protocol::ERR; //should never happen?
        }
        /*
	  arg_t *a = new arg_t(vs.seqno, proc, req);
	  insert(a);
	  while (requests.begin() != requests.end() &&
	  requests.front()->seqno == myvs.seqno) {
	  rep = execute(a->procno, a->req);
	  breakpoint1();
	  last_myvs = myvs;
	  myvs.seqno++;
	  requests.pop_front();
	  }
         */
      }
    }

  } else {

    wfs_log(WFS_DBG_1,"error: node is not stable\n");

    ret = rsm_protocol::BUSY;
  }

  //release:
  wfs_log(WFS_DBG_3,"doinvoke: invoke done\n");
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return ret;
}

#define MAX_LOG_TRANS_UNIT 1000
/**
  RPC handler: Send back the checkpoint state to the caller
  if there is any
 */
rsm_protocol::status
rsm::cptransferreq(std::string src, viewstamp start, rsm_record &r)
{
  int ret = rsm_protocol::ERR;
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  r.vs = rlog->last_cp();

  wfs_log(WFS_DBG_3,"cptransferreq from %s start %d,%d last_cp %d,%d\n",
      src.c_str(), start.vid, start.seqno,
      r.vs.vid, r.vs.seqno);

  if (start < rlog->last_cp()) {
    if (rlog->load_checkpoint(r)) {
      r.cp = true;
      ret = rsm_protocol::OK;
    }
  }
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return ret;
}

/**
 * RPC handler: Send back the missing log entries to the caller
 */
rsm_protocol::status
rsm::logtransferreq(std::string src, viewstamp start, viewstamp end,
    std::vector<rsm_record> &r)
{
  int ret = rsm_protocol::OK;

  wfs_log(WFS_DBG_3,"logtransferreq from %s vs start (%d,%d) end (%d,%d)\n", src.c_str(), start.vid, start.seqno,
      end.vid, end.seqno);

  assert(pthread_mutex_lock(&rsm_mutex)==0);

  // first see if we've missed any checkpoints
  rsm_record rec;
  if (start < rlog->last_cp()) {
    rec.vs = rlog->last_cp();

    wfs_log(WFS_DBG_2,"logtransferreq missing checkpoint %d,%d\n",
          rec.vs.vid, rec.vs.seqno);

    if (rlog->load_checkpoint(rec)) {
      rec.cp = true;
      r.push_back(rec);
    } else {
      assert(pthread_mutex_unlock(&rsm_mutex)==0);
      return rsm_protocol::ERR;
    }
  }

  rlog->reset_read();
  while (end > start) {
    if (!rlog->load_request(start, rec))
      break;
    start = rec.vs;
    r.push_back(rec);
    if (r.size() > MAX_LOG_TRANS_UNIT)
      break;
  }
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return ret;
}

/**
 * RPC handler: Send back the local node's state to the caller
 */
rsm_protocol::status
rsm::transferreq(std::string src, viewstamp vs, std::string &r)
{
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  int ret = rsm_protocol::OK;

  wfs_log(WFS_DBG_3,"transferreq from %s vs (%d,%d)\n", src.c_str(), vs.vid, vs.seqno);

  if (myvs != vs)
    ret = rsm_protocol::ERR;
  if (stf)
    r = stf->marshal_state();
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return rsm_protocol::OK;
}

/**
 * RPC handler: Send back the local node's latest viewstamp
 */
rsm_protocol::status
rsm::viewstampreq(unsigned int vid, viewstamp &r)
{
  int ret = rsm_client_protocol::OK;

  assert (pthread_mutex_lock(&rsm_mutex) == 0);

  wfs_log(WFS_DBG_3,"viewstampreq: my last vs (%d,%d) pxs stable %d vid %d\n",
        last_myvs.vid, last_myvs.seqno, pxs->is_stable(), pxs->vid());

  r = last_myvs;
  assert (pthread_mutex_unlock(&rsm_mutex) == 0);
  return ret;
}

/*
 * RPC handler: Send back all the ndoes this local knows about to client
 * so the client can switch to a different master
 * when it existing master fails
 */
rsm_client_protocol::status
rsm::client_members(int i, std::vector<std::string> &r)
{
  std::vector<std::string> m;
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  pxs->get_members(m);
  m.push_back(master);
  r = m;
  wfs_log(WFS_DBG_3,"rsm::client_members return %s\n", pxs->print_members(r).c_str());
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return rsm_protocol::OK;
}

std::string
rsm::get_master()
{
  if (!pxs->is_stable() || !stable)
    return "";

  return master;
}
bool
rsm::amimaster()
{
  assert(pthread_mutex_lock(&rsm_mutex)==0);
  bool r = get_master() == pxs->myaddr();
  assert(pthread_mutex_unlock(&rsm_mutex)==0);
  return r;
}
bool
rsm::amimaster_mx()
{
  return (get_master() == pxs->myaddr());
}
