/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// overdisk client interface.
// really the "clerk" that decides what server to talk to
// as well as sending RPCs.
// you only need one instance per host.
// assumes threading: blocking calls, uses pthreads locks, &c.
// intended to be called from a fuse client, but doesn't actually
// know anything about fuse.

#ifndef __WFS_CLIENT_H__ 
#define __WFS_CLIENT_H__ 1

#include "wfs_protocol.h"
#include "wfs_server.h"

#include <ext/hash_map>
#include <host.h>
#include <slice_manager.h>
#include <string>
#include <wfs_cache.h>
#include <wfs_client_manager.h>
#include <wfs_node.h>
#include <wfs_storage.h>
#include <vivaldi.h>

using __gnu_cxx::hash_map;

const time_t WFS_CLIENT_TIMEOUT_CACHE_SEC = 10;

class wfs_client {

 protected:

  template<class R>
    wfs_protocol::status repeated_call( wfs_file_handle fh, 
					unsigned int proc,
					marshall &req,
					R & r );

  virtual wfs_protocol::status readfile_from( sockaddr_in addr, 
					      wfs_file_handle, 
					      wfs_version version, 
					      unsigned long long off, 
					      unsigned int n, 
					      const std::string &user,
					      wfs_protocol::readinfo &info,
					      bool from_client=false);

  slice_manager *_slices;

  wfs_node _me;
  vivaldi *_viv;
  wfs_client_manager *_clients;

  wfs_server *_local_server;

  hash_map< uint32_t, time_t > _last_timeouts;

  wfs_protocol::status lookup( wfs_file_handle, std::string name, 
			       const std::string &user,
			       wfs_protocol::lookupinfo &, 
			       host *client = NULL );

  inline bool is_local( wfs_node file_node ) {
    return (file_node.id == _me.id && _local_server);
  }

 public:
  wfs_client( wfs_node me, slice_manager *slices, vivaldi *viv, 
	      wfs_client_manager *clients );
  virtual ~wfs_client() {}

  virtual bool isfile(wfs_file_handle);
  virtual bool isdir(wfs_file_handle);
  virtual bool islink(wfs_file_handle);

  void set_local_server( wfs_server *server ) { _local_server = server; }
  bool has_local_server() { return _local_server != NULL; }

  inline wfs_node get_primary( wfs_file_id fid ) {
    host h;
    slice_manager::ret_code r = 
      _slices->get_slice_primary( wfs_slice::get_slice_id(fid), h, false );
    if( r == slice_manager::SM_SUCCESS ) {
      return make_node(h);
    } else {
      wfs_log( WFS_DBG_1, 
	       "Get primary failed for fid %016"PRIx64"; failing.\n", fid );
      wfs_node x;
      x.id = 0;
      return x;
    }
  }

  // send an RPC to wfs_server
  virtual wfs_protocol::status getfile( wfs_file_handle &, 
					const std::string &user,
					wfs_protocol::fileinfo &,
					host *client = NULL );
  virtual wfs_protocol::status getfiles( std::vector<wfs_file_handle> &, 
					 const std::string &user,
					 std::vector<wfs_protocol::fileinfo> &,
					 host *client = NULL );
  virtual wfs_protocol::status getdir( wfs_file_handle &, 
					const std::string &user,
				       wfs_protocol::dirinfo &,
				       host *client = NULL );
  virtual wfs_protocol::status readdir( wfs_file_handle, 
					unsigned long long ind, unsigned int n,
					const std::string &user,
					std::vector<wfs_storage::dirent> &e,
					host *client = NULL);
  virtual wfs_protocol::status readfile( wfs_file_handle, wfs_version version, 
					 unsigned long long off, 
					 unsigned int n, 
					 const std::string &user,
					 std::string &buf );
  virtual wfs_protocol::status writefile( wfs_file_handle, wfs_version version, 
					  unsigned long long off, 
					  const std::string &buf,
					const std::string &user,
					  wfs_version &new_version,
					  host *client = NULL);
  wfs_protocol::status writefile_batch( wfs_file_handle fh, 
					std::vector<wfs_protocol::write> writes, 
					const std::string &user,
					wfs_version &new_version,
					host *client = NULL);
  virtual wfs_protocol::status lookup( wfs_file_handle, 
				       std::string name, 
				       const std::string &user,
				       wfs_file_handle &xfh,
				       host *client = NULL );
  virtual wfs_protocol::status unlink( wfs_file_handle, 
				       std::string name, 
				       bool is_dir,
				       bool dec_link,
				       const std::string &user,
				       wfs_protocol::file_and_version &fv,
				       host *client = NULL);
  virtual wfs_protocol::status create( wfs_file_handle, 
				       std::string name, 
				       const std::map<std::string,
				       wfs_permission> &permissions,
				       wfs_file_type type,
					const std::string &user,
				       wfs_file_handle &,
				       host *client = NULL);
  virtual wfs_protocol::status rename( wfs_file_handle old_parent, 
				       std::string old_name,
				       wfs_file_handle new_parent, 
				       std::string new_name,
					const std::string &user,
				       host *client = NULL);
  virtual wfs_protocol::status readlink( wfs_file_handle, wfs_version version,
					const std::string &user,
					 std::string &link );
  virtual wfs_protocol::status symlink( wfs_file_handle, 
					std::string name, 
					std::string link,
					const std::string &user,
					wfs_file_handle &,
					host *client = NULL );
  virtual wfs_protocol::status link( wfs_file_handle, 
				     wfs_file_handle,
				     std::string name,
				     bool create_file, 
				     bool link_file,
				     wfs_file_type type,
				     const std::map<std::string,wfs_permission> &permissions, 
				     const std::string &user,
				     wfs_file_handle &,
				     wfs_version &,
				     host *client = NULL );
  virtual wfs_protocol::status flush( wfs_file_handle fh, 
				      const std::string &user,
				      wfs_version &new_version,
				      host *client = NULL);
  virtual wfs_protocol::status truncate( wfs_file_handle, wfs_version version,
					 unsigned long long off,
					const std::string &user,
					 wfs_version &new_version,
					 host *client = NULL);
  virtual wfs_protocol::status chmod( wfs_file_handle fh, 
				      const std::map<std::string,
				      wfs_permission> &permissions, 
				      const std::string &user,
				      wfs_version &new_version,
				      host *client = NULL);
};

#endif // __WFS_CLIENT_H__ 
