/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


//generic asynchronous fetching for files and directories

#ifndef __WFS_FETCH_H__
#define __WFS_FETCH_H__ 1

#include <ext/hash_map>
#include <queue>
#include <vector>
#include "wfs_client.h"

//using __gnu_cxx::hash;
using __gnu_cxx::hash_map;

class wfs_cache_client;

struct req_tag {
  wfs_file_id fid;
  unsigned int bno;
  
  req_tag() {
    fid = 0;
    bno = 0;
  };
  
  req_tag(wfs_file_id _fid, 
	  unsigned int _bno) {
    fid = _fid;
    bno = _bno;
  };
    
  req_tag(wfs_file_id _fid) {
    fid = _fid;
    bno = 0;
  };
    
  bool operator==(const req_tag &r) const {
    return ((fid == r.fid) && (bno == r.bno));
  };
    
};
  
namespace __gnu_cxx {
  template<>
    class hash<req_tag> {
  public:
    size_t operator()(const req_tag &r) const {
      return 37*((unsigned int)r.fid)+r.bno;
    };
  };
};

//info for fetching request

//possible states for requests: 
// if not in the pending table, then the prefetcher
// doesn't know about the request or a non-prefetching thread
// has claimed it. Best bet is to check the cache. 
  
// if in the prefetching table, but thread field in the request info is NULL,
// then the request is waiting to be started
  
// if in the prefetching table and the thread field is filled in,
// then the request is either currently being processed or was processed
  
// check on the working request field of the thread info to find out if the 
// thread is done processing the request

// The one case where it is possible for for two requests to go 
// out for the same piece of data is if a non-prefetching thread 
// claims a request by taking it out of the prefetching queue,
// then some other thread adds it and the prefetcher starts working
// on the new request before the first non-prefetching thread finishes.

struct req_info {
  wfs_file_handle fh;
  std::string user;
  wfs_version version;
  unsigned int bno;

  req_info() { };
  
  req_info(wfs_file_handle _fh,
	   unsigned int _bno,
	   wfs_version _version,
	   std::string _user) {
    fh = _fh;
    user = _user;
    bno = _bno;
    version = _version;
  };

  req_info(wfs_file_handle _fh,
	   std::string _user) {
    fh = _fh;
    user = _user;
  };
};

//abstract base class for multi-threaded requests to the server
//the prefetching classes (wfs_dirprefetch, wfs_fileprefetch) derive from this as 
//well as the class used to make block requests in parallel to the server 
//(wfs_filefetch)

class wfs_fetch {
 protected:
  wfs_client _c;
  wfs_cache_client *_cc;
  wfs_cache *_cache;

  //queue of requests
  std::queue<req_tag> _queue;
  pthread_mutex_t _q_lock;
  pthread_cond_t _q_wait;
  
  //requests are in this hash table until they have been finished and 
  //put into the cache
  hash_map<req_tag, req_info*> _pending;
  //this lock protects all of the requests in the map as well as the map itself
  pthread_mutex_t _p_lock;

 public:
  wfs_fetch(wfs_cache*, wfs_cache_client*, wfs_node, slice_manager*, vivaldi*, 
	    wfs_client_manager*); 
  virtual ~wfs_fetch();

  //adds a thread to the thread pool
  void start(int, pthread_t[], const pthread_attr_t*);
  //code to be run in the fetch thread
  virtual void fetcher() = 0;
  
};

//class that fetches blocks from the server
//used by the cache client to send block requests in parallel
//when a read file request comes in from the server
class wfs_filefetch : public wfs_fetch {
 protected:
  struct file_req : req_info {
    wfs_protocol::status ret;

  file_req(wfs_file_handle _fh, unsigned int _bno, wfs_version _version,
	   std::string _user)
    : req_info(_fh, _bno, _version, _user) 
      {
	ret = wfs_protocol::OK;
      };
    
    ~file_req() {
    };

  };

 public:
  
  wfs_filefetch(wfs_cache*, wfs_cache_client*, wfs_node, slice_manager*, vivaldi*, 
	    wfs_client_manager*); 
  
  virtual ~wfs_filefetch(){ };

  //add the block to the queue to be fetched
  bool add(wfs_file_handle, unsigned int, wfs_version, std::string);
  //wait until the block has been fetched
  wfs_protocol::status get(wfs_file_handle, unsigned int);
  virtual void fetcher();
};
#endif
