/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


// unmarshall RPCs from overfuse/wfs_client and hand them to wfs_server
// by defaults listens on port 2273, where overfuse expects it

#include <arpa/inet.h>
#include <host_marshall.h>
#include <method_thread.h>
#include <rpc.h>
#include <signal.h>
#include <wfs_cent_cserver.h>
#include <wfs_client_manager.h>
#include <wfs_maint_slices.h>
#include <wfs_metadata.h>
#include <wfs_misc.h>
#include <wfs_storage.h>

#include "wfs_options.h"
#include "wfs_server.h"

class cserver_loop {

  const char *_my_addr;
  char *_my_port;
  std::string _my_site_id;
  slice_manager *_slices;
  wfs_client_manager *_clients;
  vivaldi *_viv;
  int _ping_period;

public:

  cserver_loop( const char *addr, char *my_port, std::string my_site_id,
		vivaldi *v, int ping_period, slice_manager *slices,
		wfs_client_manager *clients ) :
    _my_addr(addr), _my_port(my_port), _my_site_id(my_site_id),
    _slices(slices), _clients(clients), _viv(v), _ping_period(ping_period) {

  }

  void loop() {

    // my host
    host h( _my_addr, _my_port, _my_site_id );

    // sleep( random() % 20 );
    while(1) {

      // just ping a random node on every iteration
      std::vector<host> hosts;
      _slices->get_all_active_hosts(hosts);
      if( hosts.size() == 0 || (hosts.size() == 1 && hosts[0] == h) ) {
	sleep(_ping_period);
	continue;
      }
      host rand_host;
      do {
	rand_host = hosts[random()%hosts.size()];
      } while( rand_host == h );

      double before = 0, after;
      if( wfs_dbg_block(1) ) {
	before = vivaldi::gettime();
      }

      wfs_node node = make_node(rand_host);
      rpcc *client = _clients->get_client(node);
      if( client == NULL ) {
	sleep(_ping_period);
	continue;
      }
      host dummy_host;
      client->call( wcc_protocol::ping, h, dummy_host, rpcc::to(10000) );
      _clients->done_with_client(node);

      if( wfs_dbg_block(WFS_DBG_1) ) {

        after = vivaldi::gettime();
        wfs_log(WFS_DBG_1, "me %s:%s vivaldi %s lat %.3f us to %s:%d\n",
            _my_addr, _my_port,
            (_viv->my_coords()).print_coord().c_str(),
            after-before, inet_ntoa( rand_host.sin().sin_addr ),
            ntohs( rand_host.port) );
        wfs_log(WFS_DBG_1, "me  %s:%s wfs_server vivaldi info %s\n", _my_addr,
            _my_port, _viv->print_allcoords().c_str());
      }
      sleep(_ping_period);
    }

  }

};

wfs_server *init_wfs_server( wfs_options *opts, vivaldi *viv,
			     slice_manager *slices,
			     wfs_client_manager *clients ) {

  // TODO: clean up messy memory management (lots of 'new's here with no
  // corresponding 'delete's.  Won't matter unless there are multiple servers.

  rpcs *server = new rpcs( htons(opts->port) );
  server->set_vivaldi(viv);
  //server->setlossy(true);

  wfs_perfile_locker *locker = new wfs_perfile_locker();;

  wfs_storage *storage = new wfs_storage(opts->dir);

  // make sure there's a slice for me
  rep_policy pol( WFS_REP_LEVEL_DEFAULT, rep_policy::STRICT, opts->me.site_id,
		  WFS_REP_LEVEL_DEFAULT );
  wfs_file_id fid;
  host tryhost;
  slice_manager::ret_code r = slices->get_new_fid( pol, fid, tryhost, true );
  wfs_log(WFS_DBG_1, "New fid -> %d, new fid %016"PRIx64"\n", r, fid );
  assert( r == slice_manager::SM_SUCCESS || r == slice_manager::SM_SA_INC );

  // hack for planetlab, because sites don't always have great connectivity
  // between the nodes
  pol.rep_num = 1;
  pol.site_num = 1;
  r = slices->get_new_fid( pol, fid, tryhost, true );
  wfs_log(WFS_DBG_1, "New fid -> %d, new fid %016"PRIx64"\n", r, fid );
  assert( r == slice_manager::SM_SUCCESS || r == slice_manager::SM_SA_INC );

#ifdef MAIL_HACK
  sleep(15); // so we could learn about other nodes and get a balanced maint
  // now to make things easier, make sure to lock the slice
  // corresponding to your node id.
  /**
  pol.rep_num = 1;
  pol.site_num = 1;
  */
  //pol.site_num = 2;
  //pol.site_num = 1;
  wfs_slice_id sid = wfs_slice::get_slice_id(node_id_to_file_id(opts->me.id));
  r = slices->get_new_fid( pol, fid, tryhost, true, sid );
  wfs_log(WFS_DBG_1, "New fid -> %d, new fid %016"PRIx64"\n", r, fid );
  assert( r == slice_manager::SM_SUCCESS || r == slice_manager::SM_SA_INC );

  pol.rep_num = 1;
  pol.site_num = 1;
  r = slices->get_new_fid( pol, fid, tryhost, true );
  wfs_log(WFS_DBG_1, "New fid -> %d, new fid %016"PRIx64"\n", r, fid );
  assert( r == slice_manager::SM_SUCCESS || r == slice_manager::SM_SA_INC );
#endif


  wfs_maint_slices *maint = new wfs_maint_slices( opts->me, server, opts->dir,
						  locker, storage, clients,
						  opts->maint_timer );

  maint->start_maint( slices );
  //maint->start_maint( slices, false );

  wfs_server *ods = new wfs_server( opts->dir, locker, maint, storage,
				    clients, slices );

  // Methods registered with ident_reg can be called with non-host
  // credentials and verify the std::string they are called with
  // (if you're using secure channels)
  server->ident_reg(wfs_protocol::getfile, ods, &wfs_server::getfile);
  server->ident_reg(wfs_protocol::getdir, ods, &wfs_server::getdir);
  server->ident_reg(wfs_protocol::readdir, ods, &wfs_server::readdir);
  server->ident_reg(wfs_protocol::readfile, ods, &wfs_server::readfile);
  server->ident_reg(wfs_protocol::writefile, ods, &wfs_server::writefile);
  server->ident_reg(wfs_protocol::writefile_batch, ods,
		    &wfs_server::writefile_batch);
  server->ident_reg(wfs_protocol::lookup, ods, &wfs_server::lookup);
  server->ident_reg(wfs_protocol::create_dirent, ods,
		    &wfs_server::create_dirent);
  server->ident_reg(wfs_protocol::create_file, ods, &wfs_server::create_file);
  server->ident_reg(wfs_protocol::truncate, ods, &wfs_server::truncate);
  server->ident_reg(wfs_protocol::unlink, ods, &wfs_server::unlink);
  server->ident_reg(wfs_protocol::rename, ods, &wfs_server::rename);
  server->ident_reg(wfs_protocol::chmod, ods, &wfs_server::chmod);
  server->ident_reg(wfs_protocol::add_link, ods, &wfs_server::add_link);

  server->reg(wfs_protocol::addcacheloc, ods, &wfs_server::addcacheloc);
  server->reg(wfs_protocol::getcacheloc, ods, &wfs_server::getcacheloc);
  server->reg(wfs_protocol::renew_lease_batch, ods,
	      &wfs_server::renew_lease_batch);

  server->reg(wcc_protocol::ping, ods, &wfs_server::pingresponder);

  cserver_loop *looper = new cserver_loop( opts->ip.c_str(), opts->port_char,
					   opts->site_id,
					   viv, 1/*opts->ping_period*/,
					   slices, clients );
  method_thread( looper, true, &cserver_loop::loop );

  return ods;

}
