/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <chash.h>
#include <dirent.h>
#include <fcntl.h>
#include <map>
#include <method_thread.h>
#include <set>
#include <sys/stat.h>
#include <wfs_file.h>
#include <wfs_misc.h>
#include <wfs_protocol.h>
#include <wfs_metadata.h>

#include "wfs_maint_dirdisk.h"

wfs_maint_dirdisk::wfs_maint_dirdisk( wfs_node me, rpcs *server,
				      std::string dir,
				      wfs_perfile_locker *locker,
				      wfs_storage *storage,
				      wfs_client_manager *clients,
				      int maint_timer ) :
  wfs_maint(me), _server_dir(dir), _locker(locker), _nodes(NULL),
  _clients(clients), _repair_queue(NULL), _repair_queue_tail(NULL),
  _maint_started(false), _fh_needs_cues(true),
  MAINT_LOOP_TIMER(maint_timer), _storage(storage) {

  assert( pthread_mutex_init( &_repair_lock, NULL ) == 0 );
  assert( pthread_cond_init( &_repair_cond, NULL ) == 0 );

  // register self for whatever RPCs will be necessary
  server->reg( wfs_maint_protocol::md_exchange, this,
	       &wfs_maint_dirdisk::exchange_metadata );
  server->reg( wfs_maint_protocol::pull_file_data, this,
	       &wfs_maint_dirdisk::pull_file_data );
  server->reg( wfs_maint_protocol::push_file_data, this,
	       &wfs_maint_dirdisk::push_file_data );

}

wfs_maint_dirdisk::~wfs_maint_dirdisk() {
  pthread_mutex_destroy( &_repair_lock );
  pthread_cond_destroy( &_repair_cond );
}

#define WFS_MAINT_SIZE 100*1024
//#define WFS_MAINT_SIZE 5*1024*1024

void wfs_maint_dirdisk::do_repair( repair_data *d, bool exit, rpcc::TO to ) {

  wfs_log(WFS_DBG_1, "%016"PRIx64": repair of %016"PRIx64" to node %016"PRIx64" starting\n",
      _me.id, d->fid, d->node.id );

  wfs_maint_protocol::status s;
  wfs_maint_protocol::file_data file_data;

  // make sure to chunk up the reads so they fit in RPCs.
  // TODO: play with the chunk size?  is 5M too big, if lots of things are
  // going in parallel?

  // TODO: remember that a node had a timeout recently and avoid
  // sending repairs to it for a while?
  if( d->pull ) {

    unsigned long long off = 0;

    do {

      rpcc *client = _clients->get_client( d->node, to );
      if( client == NULL ) {
        s = rpc_const::timeout_failure;
      } else {
        s = client->call( wfs_maint_protocol::pull_file_data,
            d->fid, d->explicit_version,
            off, WFS_MAINT_SIZE, file_data, to );
        _clients->done_with_client(d->node);
      }
      if( s == wfs_maint_protocol::OK ) {
        wfs_log(WFS_DBG_2, "%016"PRIx64": Got %d bytes of data for fid %016"PRIx64", version %d.\n",
            _me.id, (int)file_data.data.length(), d->fid,
            file_data.version.version_num);
        assert (d->explicit_version == 0
            || file_data.version.version_num == d->explicit_version);
        write_file_data( d->fid, off, file_data.data, file_data.perm_cues,
            file_data.version, file_data.permissions,
            file_data.atime, file_data.mtime,
            file_data.num_links );
      } else {
        wfs_log(WFS_DBG_1, "%016"PRIx64": Repair of %016"PRIx64" was not ok: %d\n", _me.id,
              d->fid, s );
        break;
      }
      off += file_data.data.size();

    } while( file_data.data.size() == WFS_MAINT_SIZE );

    _locker->mark_stale( file_id_to_file_handle(d->fid), false );

  } else {

    unsigned long long off = 0;

    do {

      int32_t size = WFS_MAINT_SIZE;
      if( d->append_data != "" ) {
        size = -1;
        file_data.data = d->append_data;
      }
      read_file_data( d->fid, d->explicit_version, off, size,
          file_data.data, file_data.perm_cues,
          file_data.version, file_data.permissions,
          file_data.atime, file_data.mtime,
          file_data.num_links);

      // don't send all the data if the other side will just delete
      // TODO: if just the metadata changed, there's no need to send
      // all the data
      if( file_data.num_links == 0 ) {
        file_data.data = "";
      }

      int dummy;
      rpcc *client = _clients->get_client( d->node, to );
      if( client == NULL ) {
        s = rpc_const::timeout_failure;
      } else {
        s = client->call( wfs_maint_protocol::push_file_data,
            d->fid, off, file_data,
            (uint16_t) (d->append_data != ""),
            dummy, to );
        _clients->done_with_client(d->node);
      }
      if( s != wfs_maint_protocol::OK ) {
        break;
      }

      off += file_data.data.size();

    } while( file_data.data.size() == WFS_MAINT_SIZE );

    if( s != wfs_maint_protocol::OK ) {
      wfs_log(WFS_DBG_1, "%016"PRIx64": data push failed (%d)!\n", _me.id, s );
    }
  }

  wfs_log(WFS_DBG_1, "%016"PRIx64": repair of %016"PRIx64" to node %016"PRIx64" complete\n",
        _me.id, d->fid, d->node.id );

  if( exit ) {

    // if sync_left is negative, that means the caller already exited,
    // and if we're the last one left we should destroy everything
    // and clean up memory

    pthread_mutex_lock( d->sync_lock );
    d->statuses->push_back(s);
    if( (*d->sync_left) < 0 ) {
      (*(d->sync_left))++;
      if( *(d->sync_left) == 0 ) {
        pthread_mutex_unlock( d->sync_lock );
        pthread_mutex_destroy( d->sync_lock );
        pthread_cond_destroy( d->sync_cond );
        delete d->statuses;
        delete d->sync_left;
        delete d->sync_lock;
        delete d->sync_cond;
      }
    } else {
      (*(d->sync_left))--;
      pthread_cond_signal( d->sync_cond );
      pthread_mutex_unlock( d->sync_lock );
    }
    d->append_data = "";
    delete d;
    pthread_exit( (void *) s );
  }
}

void wfs_maint_dirdisk::repair_loop() {

  while(1) {

    pthread_mutex_lock( &_repair_lock );

    while( _repair_queue == NULL ) {
      pthread_cond_wait( &_repair_cond, &_repair_lock );
    }

    repair_data *d = _repair_queue;

    if( wfs_dbg_block(WFS_DBG_1) ) {
      if( d->pull ) {
        wfs_log(WFS_DBG_1, "%016"PRIx64": pulling repairing %016"PRIx64" from node %016"PRIx64"\n",
            _me.id, d->fid, d->node.id );
      } else {
        wfs_log(WFS_DBG_1,"%016"PRIx64": pushing repairing %016"PRIx64" to node %016"PRIx64"\n",
            _me.id, d->fid, d->node.id );
      }
      if (d->explicit_version != 0) {
        wfs_log(WFS_DBG_1,"Explicit version: %d\n", d->explicit_version);
      }
    }

    _repair_queue = d->next;
    if( _repair_queue == NULL ) {
      _repair_queue_tail = NULL;
    }

    // if we're pushing, we should allow re-enqueuing of the
    // same data as soon as we start to read it off disk, since
    // there could be important new data we'd be missing
    if( !d->pull ) {
      _repair_map[d->fid][d->node.id][d->explicit_version] = NULL;
    }

    pthread_mutex_unlock( &_repair_lock );

    // perform repair
    do_repair( d, false, rpcc::to(5000) );

    if( d->pull ) {
      // if we're pulling, don't allow re-enqueuing until after the
      // data is pulled down
      pthread_mutex_lock( &_repair_lock );
      _repair_map[d->fid][d->node.id][d->explicit_version] = NULL;
      pthread_mutex_unlock( &_repair_lock );
    }

    d->append_data = "";
    delete d;

  }

}

void wfs_maint_dirdisk::add_to_repair_queue( wfs_file_id fid, wfs_node node,
					     bool pull, bool add_to_front,
					     int32_t explicit_version) {

  repair_data *d = new repair_data();
  d->fid = fid;
  d->node = node;
  d->pull = pull;
  d->explicit_version = explicit_version;
  d->next = NULL;
  d->front = add_to_front;
  d->append_data = "";
  d->sync_lock = NULL;
  d->sync_cond = NULL;
  d->sync_left = NULL;
  d->statuses = NULL;

  pthread_mutex_lock( &_repair_lock );

  // already going to be repaired
  // TODO: if it's being added to the front of the queue, should we move
  // the existing entry to the front of the queue?
  if( _repair_map[fid][node.id][explicit_version] != NULL ) {
    pthread_mutex_unlock( &_repair_lock );
    delete d;
    return;
  }

  // if we're getting it from someone else, we know we're out of date
  // TODO: before a push, tell the other node their copy is stale??
  if( pull ) {
    _locker->mark_stale( file_id_to_file_handle(fid), true );
  }

  if( _repair_queue == NULL ) {
    _repair_queue = d;
    _repair_queue_tail = d;
  } else if( add_to_front ) {
    // put it after everyone who have also been added to the front
    repair_data *front_end = _repair_queue;
    repair_data *last_front = NULL;
    while( front_end != NULL && front_end->front ) {
      last_front = front_end;
      front_end = front_end->next;
    }
    d->next = front_end;
    if( last_front == NULL ) {
      _repair_queue = d;
    } else {
      last_front->next = d;
    }
    if( front_end == NULL ) {
      _repair_queue_tail = d;
    }
  } else {
    _repair_queue_tail->next = d;
    _repair_queue_tail = d;
  }

  wfs_log(WFS_DBG_1, "%016"PRIx64": Added file %016"PRIx64"/node %016"PRIx64" to the queue for repair (ev=%d)\n",
	    _me.id, fid, node.id, explicit_version );

  _repair_map[fid][node.id][explicit_version] = d;

  pthread_cond_signal( &_repair_cond );

  pthread_mutex_unlock( &_repair_lock );

}

void wfs_maint_dirdisk::get_replicas( wfs_file_handle fh,
				      std::vector<wfs_node> &reps ) {

  int rep_level = fhreplevel(fh);
  std::set<std::string> sites;
  wfs_node last_node;
  last_node.id = 0;
  for( int i = 1; i <= rep_level; i++ ) {
    wfs_node next_node = _nodes->list()->successor( fh.fid, i );

    bool cont = false;

    // Note that site-replication stuff only works right as long as
    // the node that originally got the file is still the responsible
    // node.  If responsibility changes and the responsible node's site
    // is different, then the backups will switch configuration
    // based on what the new site is.  This may not be desirable.  Fixing it
    // is a hassle though, since then the site would probably need to be
    // pushed around in the file handle and taken into account when
    // doing consistent hash lookups on the file id.

    // one site replication
    if( fh.get_cue(WFS_CUE_REP_SITES).val == 1 ) {
      if( next_node.site_id != _me.site_id ) {
	rep_level++;
	cont = true;
      }
    }

    // TODO: This isn't quite the right use of RepSites, but it'll do
    // until we have the cserver pick the backups

    // multi-site replication
    if( fh.get_cue(WFS_CUE_REP_SITES).val > i ) {
      if( sites.find(next_node.site_id) != sites.end() ) {
	rep_level++;
	cont = true;
      } else {
	sites.insert( next_node.site_id );
      }
    }

    if( next_node.id == last_node.id ) {
      break;
    }

    last_node = next_node;

    if( cont ) {
      continue;
    }

    reps.push_back( next_node );
  }

}

void wfs_maint_dirdisk::immediate_replica_enqueue( wfs_file_handle fh ) {

  if( !_maint_started ) {
    // maintenance is turned off . . .
    return;
  }

  wfs_log(WFS_DBG_1, "%016"PRIx64": Immediately enqueuing file %016"PRIx64"\n", _me.id,
	    fh.fid  );

  // NOTE: assumes fid is already locked in the locker

  // figure out all the successors and sync
  std::vector<wfs_node> reps;
  get_replicas( fh, reps );

  wfs_log(WFS_DBG_2, "%016"PRIx64": Got replicas for file %016"PRIx64"\n", _me.id, fh.fid );

  for( unsigned i = 0; i < reps.size(); i++ ) {
    wfs_node succ = reps[i];
    if( succ.id != _me.id ) {
      wfs_log(WFS_DBG_2, "%016"PRIx64": Enqueueing repair to node %016"PRIx64" for file %016"PRIx64"\n", _me.id, succ.id, fh.fid );
      add_to_repair_queue( fh.fid, succ, false, true );
    }
  }

  wfs_log(WFS_DBG_2,"%016"PRIx64": Finished enqueuing file %016"PRIx64"\n", _me.id, fh.fid );

}

wfs_maint_protocol::status wfs_maint_dirdisk::immediate_replica_sync( wfs_file_handle fh,
								      wfs_file_handle other_fh,
								      std::string append_data ) {

  wfs_log(WFS_DBG_1, "%016"PRIx64": Immediately syncing file %016"PRIx64"\n", _me.id, fh.fid );

  // NOTE: Assume file is locked already

  if( !_maint_started ) {
    // maintenance is turned off . . .
    return wfs_maint_protocol::OK;
  }

  // figure out all the successors and sync
  std::vector<wfs_node> reps;
  get_replicas( fh, reps );

  if( wfs_dbg_block(WFS_DBG_1) ) {
    wfs_log(WFS_DBG_1, "Replica set for file %016"PRIx64" is ", fh.fid );
    for( unsigned i = 0; i < reps.size(); i++ ) {
      wfs_log(WFS_DBG_1_NOHEADER, "\t\t\t%016"PRIx64" ", reps[i].id );
    }
  }

  // unlock the file for the duration of the repairs
  _locker->unlock_file( fh );
  if( other_fh.fid != 0 && other_fh.fid != fh.fid ) {
    _locker->unlock_file(other_fh);
  }

  pthread_mutex_t *sync_lock = new pthread_mutex_t();
  pthread_mutex_init( sync_lock, NULL );
  pthread_cond_t *sync_cond = new pthread_cond_t();
  pthread_cond_init( sync_cond, NULL );
  std::vector<wfs_maint_protocol::status> *statuses =
    new std::vector<wfs_maint_protocol::status>;
  int *sync_left = new int();
  pthread_mutex_lock( sync_lock );
  *sync_left = 0;
  int total = 0;

  for( unsigned i = 0; i < reps.size(); i++ ) {
    wfs_node succ = reps[i];
    if( succ.id != _me.id ) {
      repair_data *d = new repair_data();
      d->fid = fh.fid;
      d->node = succ;
      d->pull = false;
      d->explicit_version = 0; // xavid: This function is only called for new versions
      d->next = NULL;
      d->append_data = append_data;
      d->sync_lock = sync_lock;
      d->sync_cond = sync_cond;
      d->sync_left = sync_left;
      d->statuses = statuses;
      (*sync_left)++;
      total++;

      pthread_t pth = method_thread( this, true,
				     &wfs_maint_dirdisk::do_repair, d, true,
				     fhto(fh) );
      assert(pth);

    }
  }

  wfs_maint_protocol::status final_status = wfs_maint_protocol::OK;

  while( (*sync_left) > 0 &&
	 (total - (*sync_left)) < (fhsynclevel(fh)-1) ) {

    wfs_log(WFS_DBG_1, "%016"PRIx64": backing up file %016"PRIx64", waiting for %d threads\n",
	      _me.id, fh.fid, fhsynclevel(fh) - 1 - (total - (*sync_left)) );

    pthread_cond_wait( sync_cond, sync_lock );

  }

  for( unsigned i = 0; i < statuses->size(); i++ ) {
    wfs_maint_protocol::status s = (*statuses)[i];
    if( s != wfs_maint_protocol::OK ) {
      final_status = s;
    }
  }

  bool destroy =  (*sync_left == 0);
  *sync_left = -(*sync_left); // signal that we've exited

  pthread_mutex_unlock( sync_lock );

  if( destroy ) {
    pthread_mutex_destroy( sync_lock );
    pthread_cond_destroy( sync_cond );
    delete statuses;
    delete sync_left;
    delete sync_lock;
    delete sync_cond;
  }

  if( other_fh.fid == fh.fid || other_fh.fid == 0 ) {
    _locker->lock_file(fh);
  } else if( other_fh.fid < fh.fid ) {
    assert( other_fh.fid != 0 );
    _locker->lock_file(other_fh);
    _locker->lock_file(fh);
  } else {
    assert( other_fh.fid != 0 );
    _locker->lock_file(fh);
    _locker->lock_file(other_fh);
  }

  wfs_log(WFS_DBG_1, "%016"PRIx64": Finished immediately syncing file %016"PRIx64"\n",
	    _me.id, fh.fid );

  return final_status;

}

void wfs_maint_dirdisk::build_metadata_list( wfs_node_id node_id,
					     std::vector<wfs_maint_protocol::metadata> &md ) {

  wfs_log(WFS_DBG_1, "%016"PRIx64": building metadata list against node %016"PRIx64"\n",
	    _me.id, node_id );

  // find all the data objects for which node and I are both in the
  // replica list

  DIR *dir = opendir( _server_dir.c_str() );

  if( dir != NULL ) {

    struct dirent *entry;
    while( (entry = readdir(dir)) ) {
      wfs_file_id fid;
      uint32_t fid_1;
      uint32_t fid_2;
      char dummy[20];
      // ignore any dot extensions of a file
      if( sscanf( entry->d_name, "%08"SCNx32"%08"SCNx32".%s",
		  &fid_1, &fid_2, dummy ) != 3 &&
	  sscanf( entry->d_name, "%08"SCNx32"%08"SCNx32,
		  &fid_1, &fid_2 ) == 2 ) {

	fid = two_ints_to_file_id( fid_1, fid_2 );

	wfs_file_handle fh;
	fh.fid = fid;
	wfs_metadata *md_file = NULL;
	// for slice-based maintenance, you only need to read the meta-data
	// if the file is in common
	if( _fh_needs_cues ) {
	  _locker->lock_file( fh );
	  wfs_metadata *md_file = _storage->get_md(fid);
	  fh.set_perm_cues( md_file->perm_cues );
	}

	// are we both in the replica list?
	//int rep_level = fhreplevel(fh);
	std::vector<wfs_node> reps;
	get_replicas( fh, reps );
	//wfs_node_id fid_node = file_id_to_node_id( fid );

	// make sure this node is indeed on the list of sucessors
	bool is_rep = false;
	bool me_rep = false;
	for( unsigned i = 0; i < reps.size(); i++ ) {
	  if( reps[i].id == node_id ) {
	    is_rep = true;
	  } else if( reps[i].id == _me.id ) {
	    me_rep = true;
	  }
	}

	if( is_rep && me_rep ) {

	  wfs_maint_protocol::metadata curr_md;
	  curr_md.fh = fh;

	  if( !_fh_needs_cues ) {
	    _locker->lock_file( fh );
	    md_file = _storage->get_md(fid);
	    fh.set_perm_cues( md_file->perm_cues );
	  }

	  // set the version number
	  curr_md.version = md_file->version;

	  if( wfs_isdir(fh) ) {
	    curr_md.checksum = _storage->file_checksum(fh);
	  } else {
	    curr_md.checksum = 0;
	  }

	  wfs_log(WFS_DBG_1, "%016"PRIx64": Pushing %016"PRIx64" onto the md vector in build_metadata_list against node %016"PRIx64"\n",
		    _me.id, fid, node_id );
	  md.push_back( curr_md );

	  if( !_fh_needs_cues ) {
	    _storage->release_md( md_file );
	    _locker->unlock_file( fh );
	  }

	}

	if( _fh_needs_cues ) {
	  _storage->release_md( md_file );
	  _locker->unlock_file( fh );
	}

      }
    }

    closedir(dir);

  } else {
    wfs_log(WFS_DBG_OFF, "Couldn't open server directory %s\n",
	     _server_dir.c_str() );
  }

}

void wfs_maint_dirdisk::compare_metadata( wfs_node node,
					  std::vector<wfs_maint_protocol::metadata> my_md,
					  std::vector<wfs_maint_protocol::metadata> node_md ) {


  if( node.id == _me.id || node.id == 0 ) {
    return;
  }

  // see what data my neighbor has that I need, and place on repair list
  std::map<wfs_file_id, wfs_maint_protocol::metadata> my_map;

  for( unsigned i = 0; i < my_md.size(); i++ ) {
    wfs_maint_protocol::metadata md = my_md[i];
    wfs_log(WFS_DBG_1, "%016"PRIx64": Adding %016"PRIx64",%d to metadata for repair\n",
	      _me.id, md.fh.fid, md.version.version_num );
    my_map[md.fh.fid] = md;
  }

  for( unsigned i = 0; i < node_md.size(); i++ ) {

    wfs_maint_protocol::metadata md = node_md[i];

    // make sure we're both on the replicas set, otherwise we could try to sync
    // with an old member, slowing us down
    std::vector<wfs_node> reps;
    get_replicas( md.fh, reps );
    bool is_rep = false;
    bool me_rep = false;
    for( unsigned j = 0; j < reps.size(); j++ ) {
      if( reps[j].id == node.id ) {
	is_rep = true;
      } else if( reps[j].id == _me.id ) {
	me_rep = true;
      }
    }

    if( is_rep && me_rep &&
	((my_map.find(md.fh.fid) == my_map.end()) ||
	 (my_map[md.fh.fid].version < md.version) ||
	 (wfs_isdir(md.fh) &&
	  md.fh.get_cue(WFS_CUE_CONSISTENCY).val == WFS_CUE_EVENTUAL &&
	  my_map[md.fh.fid].checksum != md.checksum)) ) {
      wfs_log(WFS_DBG_1, "%016"PRIx64": Placing %016"PRIx64",%d on the repair queue (%d)\n",
                _me.id, md.fh.fid, md.version.version_num,
                (my_map.find(md.fh.fid) == my_map.end()) );
      add_to_repair_queue( md.fh.fid, node, true, false );
      // ?xavid: does it add anything to compare both?
      if (md.fh.get_cue(WFS_CUE_VERSIONING).val == WFS_CUE_VERSIONED
	  || my_map[md.fh.fid].fh.get_cue(WFS_CUE_VERSIONING).val
	     == WFS_CUE_VERSIONED) {
	// Pull any versions we missed somewhere!
	for (int32_t missver = my_map[md.fh.fid].version.version_num + 1;
	     missver < md.version.version_num; missver++) {
	  wfs_log(WFS_DBG_1, "%016"PRIx64": Also placing version %d on the repair queue\n",
		    _me.id, missver );
	  add_to_repair_queue(md.fh.fid, node, true, false, missver);
	}
      }
    }
  }

}

bool wfs_maint_dirdisk::do_metadata_exchange( wfs_node node ) {

  // ask pred for all the meta-data it knows about between his id and
  // my own, if I'm within in the replication level for that data.
  // If the versions don't match, put the data in a queue somewhere to
  // reconcile at some later date

  std::vector<wfs_maint_protocol::metadata> my_md, node_md;
  build_metadata_list( node.id, my_md );

  rpcc *client = _clients->get_client( node, rpcc::to(5000) );
  if( client == NULL ) {
    // TODO: better timeout???
    return false;
  } else {
    wfs_maint_protocol::status s = client->call(wfs_maint_protocol::md_exchange,
						_me.id, my_md, node_md,
						rpcc::to(5000));
    _clients->done_with_client(node);
    //+xavid
    if( s != wfs_maint_protocol::OK ) {
      return false;
    }
  }

  my_md.clear();
  // rebuild list, the old one could already be out of date
  build_metadata_list( node.id, my_md );

  compare_metadata( node, my_md, node_md );

  return true;

}

wfs_node wfs_maint_dirdisk::node_from_id( wfs_node_id node_id ) {
  return _nodes->list()->successor( node_id );
}

int wfs_maint_dirdisk::exchange_metadata( wfs_node_id node_id,
					  std::vector<wfs_maint_protocol::metadata> in_md,
					  std::vector<wfs_maint_protocol::metadata> &out_md ) {

  build_metadata_list( node_id, out_md );
  // fork off a thread for this, so we don't block the response?
  compare_metadata( node_from_id(node_id), out_md, in_md );

  return wfs_maint_protocol::OK;

}

int wfs_maint_dirdisk::pull_file_data( wfs_file_id fid, int32_t vernum,
				       unsigned long long offset, int32_t size,
				       wfs_maint_protocol::file_data &file_data ) {

  wfs_log(WFS_DBG_1, "%016"PRIx64": servicing pulled repair for file %016"PRIx64"\n",
	    _me.id, fid );

  return read_file_data( fid, vernum, offset, size, file_data.data,
			 file_data.perm_cues, file_data.version,
			 file_data.permissions, file_data.atime,
			 file_data.mtime, file_data.num_links );

}

int wfs_maint_dirdisk::push_file_data( wfs_file_id fid,
				       unsigned long long offset,
				       wfs_maint_protocol::file_data file_data,
				       uint16_t append,
				       int &dummy) {

  wfs_log(WFS_DBG_1, "%016"PRIx64": receiving pushed repair for file %016"PRIx64",%d append:%d, %u bytes at offset %llu\n",
	    _me.id, fid, file_data.version.version_num, append, (unsigned)file_data.data.size(), offset );

  return write_file_data( fid,
			  offset, file_data.data, file_data.perm_cues,
			  file_data.version, file_data.permissions,
			  file_data.atime, file_data.mtime,
			  file_data.num_links, append );
}

int wfs_maint_dirdisk::read_file_data( wfs_file_id fid,
				       int32_t vernum,
				       unsigned long long offset,
				       int32_t size,
				       std::string &data,
				       std::string &perm_cues,
				       wfs_version &version,
				       std::map<std::string,
				       wfs_permission> &permissions,
				       unsigned long &atime,
				       unsigned long &mtime,
				       int &num_links ) {

  wfs_file_handle fh = file_id_to_file_handle(fid);
  _locker->lock_file(fh);

  if( size >= 0 ) {
    if( !_storage->object_exists( fh, vernum ) ) {
      _locker->unlock_file(fh);
      return wfs_maint_protocol::NOENT;
    }
    wfs_version v;
    v.version_num = vernum;
    if( !_storage->readfile( fh, v, offset, size, data ) ) {
      _locker->unlock_file(fh);
      return wfs_maint_protocol::IOERR;
    }
  }

  wfs_metadata *md = _storage->get_md( fid, vernum );
  perm_cues = md->perm_cues;
  permissions = md->permissions;
  wfs_log(WFS_DBG_2, "md with vernum=%d has version.version_num=%d!\n",
	   vernum, md->version.version_num);
  version = md->version;
  num_links = md->num_links;
  _storage->release_md(md);

  struct stat sb;
  if( _storage->stat_object(fh, &sb, vernum) ) {
    atime = sb.st_atime;
    mtime = sb.st_mtime;
  }

  _locker->unlock_file(fh);
  return wfs_maint_protocol::OK;

}

int wfs_maint_dirdisk::merge_dirs( wfs_file_id fid, std::string remote_dir ) {

  std::map<std::string, wfs_file_id> on_local;
  std::vector<wfs_storage::dirent> local;
  wfs_file_handle fh = file_id_to_file_handle(fid);
  fh.set_cue( WFS_CUE_CONSISTENCY, WFS_CUE_EVENTUAL );
  if( !_storage->readdir( fh, 0, 1024*1024*1024, local ) ) {
    return wfs_maint_protocol::IOERR;
  }

  for( unsigned i = 0; i < local.size(); i++ ) {
    on_local[local[i].name] = file_handle_to_file_id(local[i].inum);
  }

  // now go through the remote directory and see if we're missing anything
  // TODO: clean this up so it doesn't depend on the on-disk data format
  char id[WFS_FID_NUM_HEX_BYTES+1];
  int j = 0;
  bool in_name = true;
  //bool updated = false;
  wfs_file_id curr_fid;
  char zname[WFS_DIRENTRY_MAX_NAMELEN+1];
  for( unsigned int i = 0; i < remote_dir.size(); i++ ) {
    char c = remote_dir[i];
    if( c == '\n' && !in_name ) {
      in_name = true;
      id[j] = '\0';
      j = -1;

      uint32_t fid_1;
      uint32_t fid_2;
      assert( sscanf( id, "%08"SCNx32"%08"SCNx32"", &fid_1, &fid_2 ) == 2 );
      curr_fid = two_ints_to_file_id( fid_1, fid_2 );
      std::string zstr(zname);
      if( on_local.find( zstr ) == on_local.end() ) {

	std::string append_data;
	if ( !_storage->link( fh, zstr, file_id_to_file_handle(curr_fid),
			      append_data ) ) {
	  return wfs_maint_protocol::IOERR;
	}
        on_local[zstr] = curr_fid;
        wfs_log(WFS_DBG_1, "%016"PRIx64": Adding missing "
            "entry (%s, %016"PRIx64") to dir %016"PRIx64"\n",
                  _me.id, zname, curr_fid, fid );

      } else if( curr_fid < on_local[zstr] ) {

	// TODO: have a change_link, instead of unlink and link
	if( !_storage->unlink( fh, zstr ) ) {
	  return wfs_maint_protocol::IOERR;
	}
	std::string append_data;
	if ( !_storage->link( fh, zstr, file_id_to_file_handle(curr_fid),
			      append_data ) ) {
	  return wfs_maint_protocol::IOERR;
	}
	on_local[zstr] = curr_fid;

      }

    } else if( c == WFS_DIRENTRY_SEP ) {
      in_name = false;
      zname[j] = '\0';
      j = -1;
    } else if( in_name ) {
      zname[j] = c;
    } else {
      id[j] = c;
    }

    j++;
  }

  return wfs_maint_protocol::OK;

}

int wfs_maint_dirdisk::write_file_data( wfs_file_id fid,
    unsigned long long offset,
    std::string data,
    std::string perm_cues,
    wfs_version version,
    const std::map<std::string,
    wfs_permission> &permissions,
    unsigned long atime,
    unsigned long mtime,
    int num_links,
    bool append ) {

  wfs_file_handle fh = file_id_to_file_handle(fid);
  fh.set_perm_cues(perm_cues);

  _locker->lock_file(fh);

  wfs_metadata *md = _storage->get_md(fid);
  wfs_version curr_version = md->version;
  int32_t use_vernum = 0;

  wfs_log(WFS_DBG_2, "writing to version %d (vs %d) of file %016"PRIx64"; versioned=%d\n",
      version.version_num, curr_version.version_num, fid,
      (int)(fh.get_cue( WFS_CUE_VERSIONING ).val == WFS_CUE_VERSIONED));

  // we already have a better version on disk, so ignore this
  // xavid: unless the file is versioned, in which case we update our
  //        saved version if necessary
  if( wfs_isfile(fid) && curr_version >= version ) {

    if (fh.get_cue(WFS_CUE_VERSIONING).val == WFS_CUE_VERSIONED) {
      // Only use the sent version if we don't have anything with
      // that version number or what we have is `lower' (comparing nodeids)
      bool exists = _storage->object_exists( fh, version.version_num );
      if( exists ) {
        _storage->release_md(md);
        md = _storage->get_md( fid, version.version_num );
        curr_version = md->version;
      }
      if (!exists || version > curr_version) {
        // Write to the old version instead of the current one
        use_vernum = version.version_num;
      } else {
        // Ours is better
        _locker->unlock_file(fh);
        _storage->release_md(md);
        return wfs_maint_protocol::OK;
      }
    } else {

      // if this write is to past the end of the file we currently have,
      // accept it
      struct stat sb;
      if( !_storage->stat_object( fh, &sb, use_vernum ) ) {
        _locker->unlock_file(fh);
        _storage->release_md(md);
        // possibly deleted already
        return wfs_maint_protocol::OK;
      }

      if( offset < (long long unsigned)sb.st_size ) {
        wfs_log(WFS_DBG_1, "Rejecting maint repair for offset %llu for file %016"PRIx64
            " is %d bytes long\n", offset, fh.fid, (int)sb.st_size );
        // Not versioned and we're trying to write to the beginning of the file
        _locker->unlock_file(fh);
        _storage->release_md(md);
        return wfs_maint_protocol::OK;
      }

    }

  } else if (curr_version.version_num < version.version_num
      && fh.get_cue( WFS_CUE_VERSIONING ).val == WFS_CUE_VERSIONED) {

    if (curr_version.version_num != 0) {
      // Archive our current as an old version before replacing
      _storage->save_old_version(fh);
    }
    // Check if we're missing any intermediate versions
    if (version.version_num - curr_version.version_num > 1) {
      std::vector<wfs_node> reps;
      get_replicas( fh, reps );

      int counter = 0;
      while( reps.size() < 2 && counter < 5 ) {
        // pseudo-hack to make sure that there's at least one replica
        // because otherwise the node could contact us before our
        // node list has caught up
        wfs_log(WFS_DBG_1, "%016"PRIx64": Waiting for a replica for %016"PRIx64
              " so we can ask for old versions\n",
              _me.id, fh.fid );
        // unlocking the file here might cause problems if the
        // meta-data changes underneath us.  but it's necessary
        // because otherwise the maintenance process might block,
        // potentially causing the replica set to never coalesce
        // (specifically in maint_slices)
        _locker->unlock_file(fh);
        sleep(MAINT_LOOP_TIMER);
        _locker->lock_file(fh);
        reps.clear();
        get_replicas( fh, reps );
        counter++;
      }

      for (int32_t missver = curr_version.version_num + 1;
      missver < version.version_num; missver++) {

        wfs_log(WFS_DBG_1, "%016"PRIx64": Asking for missing version %d of file %016"PRIx64"\n", _me.id, missver, fid );

        for( unsigned i = 0; i < reps.size(); i++ ) {
          wfs_node succ = reps[i];
          if( succ.id != _me.id ) {
            add_to_repair_queue(fid, succ, true, false, missver);
          }
        }

      }
    }
  }

  // create the file just in case
  _storage->create_new_object( fh, use_vernum );

  if( wfs_isfile(fid) ||
      fh.get_cue(WFS_CUE_CONSISTENCY).val != WFS_CUE_EVENTUAL ) {

    if( append ) {

      struct stat sb;
      if( !_storage->stat_object( fh, &sb, use_vernum ) ) {
        _locker->unlock_file(fh);
        _storage->release_md(md);
        wfs_log(WFS_DBG_OFF, "Couldn't stat file for append\n" );
        return wfs_maint_protocol::IOERR;
      }
      offset = (int) sb.st_size;
      wfs_log(WFS_DBG_1, "append mode\n" );

    } else {

      // if it's not append then we're replacing the end of this file
      // with this new write
      if( !_storage->truncate( fh, offset ) ) {
        _locker->unlock_file(fh);
        _storage->release_md(md);
        wfs_log(WFS_DBG_OFF, "truncate failed\n" );
        return wfs_maint_protocol::IOERR;
      }

    }

    std::vector<unsigned long long> offv;
    offv.push_back(offset);
    std::vector<std::string> bufv;
    bufv.push_back(data);
    if( !_storage->writefile( fh, offv, bufv ) ) {
      _locker->unlock_file(fh);
      _storage->release_md(md);
      wfs_log(WFS_DBG_OFF, "write failed\n" );
      return wfs_maint_protocol::IOERR;
    }

  } else {
    // Lax directory

    assert( offset == 0 );  // assume dirs are no bigger than the block size
    wfs_maint_protocol::status s = merge_dirs( fid, data );
    if( s != wfs_maint_protocol::OK ) {
      _locker->unlock_file(fh);
      _storage->release_md(md);
      return s;
    }

  }

  md->permissions = permissions;

  // actually want to merge the perm cues, in case we're in lax mode
  // and this file was created in two places with different cues
  std::string cues_now = md->perm_cues;
  if( cues_now.length() > 0 ) {
    wfs_file_handle now_fh;
    now_fh.set_perm_cues( cues_now );
    now_fh.set_perm_cues( perm_cues );
    perm_cues = now_fh.perm_cues_string();
  }
  md->perm_cues = perm_cues;
  md->version = version;
  md->num_links = num_links;
  md->write_to_file();
  _storage->set_and_release_md( fid, md );

  struct stat sb;
  sb.st_atime = atime;
  sb.st_mtime = mtime;
  // don't worry about errors here for now
  _storage->set_stat_object(fh, &sb, use_vernum);

  if( num_links == 0 ) {
    _storage->delete_object(fh);
  }

  _locker->unlock_file(fh);
  return wfs_maint_protocol::OK;

}

void wfs_maint_dirdisk::maint_loop() {

  // periodically contact predecessors to exchange meta-data and
  // place necessary repairs onto a queue
  wfs_node last_pred = _me;
  unsigned pred_count = 0;

  while(1) {

    wfs_node pred = _nodes->list()->predecessor( last_pred.id );
    pred_count++;
    last_pred = pred;

    // in case it wraps around:
    if( pred.id == _me.id ) {
      pred_count = 0;
      sleep( MAINT_LOOP_TIMER );
      continue;
    }

    wfs_log(WFS_DBG_1, "%016"PRIx64": exchanging with %016"PRIx64"\n", _me.id, pred.id );
    do_metadata_exchange( pred );

    // don't go backwards arbitrarily, just look at enough preds
    // to ensure you'll see all data you might be responsible for.
    if( pred_count > WFS_REP_LEVEL_MAX ) {
      last_pred = _me;
      pred_count = 0;
    }

    sleep( MAINT_LOOP_TIMER );
  }

}

void wfs_maint_dirdisk::start_maint( node_manager *nodes, bool do_maint ) {

  _nodes = nodes;
  _maint_started = true;

  if( do_maint ) {
    // check with our successor right away to get the right meta-data, and
    // mark objects as stale
    wfs_node succ = _nodes->list()->successor( _me.id );
    if( succ.id != _me.id ) {
      do_metadata_exchange( succ );
    }
    method_thread( this, true, &wfs_maint_dirdisk::maint_loop );
  }

  method_thread( this, true, &wfs_maint_dirdisk::repair_loop );

}
