/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef __COORD_H__
#define __COORD_H__

#include <string>
#include <vector>
#include <ext/hash_map>
#include "marshall.h"

#define NCOORD 2
#define USING_HT 1
#define PRED_ERR_MULT 1000.0

// Perhaps this should be made a coordinate in the sort of standard
// object-oriented sense of the word, instead of just a namespace
// for some common functions.
struct Coord {

  std::vector<double> coords;
  double ht;

  unsigned int size () const { return coords.size (); };

  void set (const Coord &coords);
  void set_coord (int i, double x);
  double get_coord (int i);
  Coord ();
  Coord (const Coord &c) : coords (c.coords), ht (c.ht), pred_err (c.pred_err) {};
  double err () const { return pred_err/PRED_ERR_MULT; };
  void update_err (double x) {  pred_err = (int)(x*PRED_ERR_MULT); };

  void set_raw_err(int e) { pred_err = e;};
  int raw_err () const { return pred_err; };
  std::string print_coord ();

  double distance_f (const Coord &c);
  double norm ();
  double plane_norm ();
  void scalar_mult (double s);
  void vector_add (const Coord &b);
  void vector_sub (const Coord &b);

  static double distance_f (const Coord &a, const Coord &n);

private:
  int pred_err;

};

class vivaldi {
  public:
	struct coord_info {
		Coord c;
		double when;
		double age;
		unsigned port; //i save this information for debugging purpose only
	};
	
	 vivaldi();
	 ~vivaldi();

	 static double gettime();
	 bool get_coords(unsigned long long k, Coord &c);
	 void update_other_coords(uint32_t ip, uint16_t port, Coord uc, double age = 0.0);
	 void update_other_lat (uint32_t ip, double lat);
	 double get_distance_to_ip (uint32_t ip, uint16_t port);

	 void update_coords(Coord uc, double ud);
	 Coord &my_coords() { return _coords; }
	 std::string print_allcoords();
  protected:
	 void update_error(double actual, double expect, double rmt_err);
	 Coord _coords;
	 pthread_mutex_t _m;
	 __gnu_cxx::hash_map< unsigned , coord_info> _info;
	 __gnu_cxx::hash_map< unsigned , double> _lat;
};

inline unmarshall &
operator>>(unmarshall &u, Coord &c)
{
  int x;
  for (int i = 0; i < NCOORD; i++) {
	 u >> x;
	 c.set_coord(i,(double)x);
  }
  u >> x;
  c.set_raw_err((int)x);
  return u;
}

inline marshall &
operator<<(marshall &m, Coord c)
{
  int x;
  for (int i = 0; i < NCOORD; i++) {
	 x = (int)c.get_coord(i);
	 m << x;
  }
  x = (int)c.raw_err();
  m << x;
  return m;
}

#endif
