#!/bin/bash -e

d1="$1"
d2="$2"
d3="$3"
p1="$4"
p2="$5"

# stat unfortunately works different on BSD than on Linux, and ls -l also
# has annoying differences, so I use this detection mechanism
if stat --format %u . > /dev/null 2>&1; then
    STAT_FORMAT="stat --format"
    MOD="%a"
    OWN="%U"
    GRP="%G"
else
    STAT_FORMAT="stat -f"
    MOD="%Lp"
    OWN="%Su"
    GRP="%Sg"
fi
SOMEONE=daemon
SOMEGROUP="bin"

function check1 {
    if cat "$1" > /dev/null 2>&1; then
	:
    else
	echo "File $1 was unreadable when it should have contained '$2'!"
	exit 1
    fi
    if [[ "`cat $1`" != "$2" ]] ; then
	echo "Contents of $1 should be '$2', not '`cat $1`'!"
	exit 1
    fi
}

function checknotreadable {
    # It looks like it exists from the dir entry, so we have to actually
    # try reading it
    if cat "$1" > /dev/null 2>&1 ; then 
	echo "File $1 is readable ('`cat $1`') when it shouldn't be!"
	exit 1
    fi
}

function checknotwritable {
    if bash -c "echo wine > '$1'" 2> /dev/null ; then 
	echo "File $1 is writable when it shouldn't be!"
	exit 1
    fi
}

function checkfails {
    if "$@" > /dev/null 2>&1 ; then 
	echo "Command '$@' succeeds when it shouldn't!"
	exit 1
    fi
}

function checkmod {
    mode="`$STAT_FORMAT $MOD \"$1\"`"
    if [[ "$mode" -ne "$2" ]] ; then
	echo "File $1 should have access mode $2, not $mode!"
	exit 1
    fi
    owner="`$STAT_FORMAT $OWN \"$1\"`"
    if [[ "$owner" != "$3" ]] ; then
	echo "File $1 should be owned by $3, not $owner!"
	exit 1
    fi
    if [[ -n "$4" ]] ; then
	group="`$STAT_FORMAT $GRP \"$1\"`"
	if [[ "$mode" != "$4" ]] ; then
	    echo "File $1 should have group $4, not $group!"
	    exit 1
	fi
    fi
}

umask 022

echo -n "Basic chmod and stat, one machine: "
echo foo > "$d1/bar"
checkmod "$d1/bar" 604 "$USER"
chmod uo+x,o-r "$d1/bar"
checkmod "$d1/bar" 701 "$USER"
chown "$SOMEONE" "$d1/bar"
checkmod "$d1/bar" 701 "$SOMEONE"
echo "OK."

echo -n "Basic chmod and stat, two machines: "
echo foo > "$d1/baz"
check1 "$d2/baz" foo
checkmod "$d2/baz" 604 "$USER"
chmod uo+x,o-r "$d2/baz"
checkmod "$d1/baz" 701 "$USER"
chown "$SOMEONE" "$d1/baz"
checkmod "$d1/baz" 701 "$SOMEONE"
checkmod "$d2/baz" 701 "$SOMEONE"
echo "OK."

echo -n "Permissions with .ident: "
echo wine > "$d2/cheese"
check1 "$d1/cheese" wine
check1 "$d1/.ident=ralph/cheese" wine
checkmod "$d1/cheese" 604 "$USER"
checkmod "$d1/.ident=xyzzyqqqqq/cheese" 004 "root"
checkmod "$d2/.ident=$SOMEONE/cheese" 004 "root"
chown "$SOMEONE" "$d1/cheese"
checkmod "$d1/cheese" 604 "$SOMEONE"
checkmod "$d2/.ident=$SOMEONE/cheese" 664 "$SOMEONE"
# Chown with ident
echo -n "(chown) "
echo pine > "$d1/fleas"
chown "root" "$d2/.ident=wombat/fleas"
checkmod "$d2/fleas" 604 "root"
checkmod "$d1/.ident=$SOMEONE/fleas" 004 "root"
checkmod "$d1/.ident=wombat/fleas" 664 "root"
# Chmod with ident
echo -n "(chmod) "
echo trine > "$d2/peas"
chmod u+x "$d2/peas"
checkmod "$d1/.ident=barnaby/peas" 004 "root"
checkmod "$d1/peas" 704 "$USER"
chmod "u+wx" "$d1/.ident=barnaby/peas"
checkmod "$d1/.ident=barnaby/peas" 334 "root"
checkmod "$d1/.ident=$SOMEGROUP/peas" 004 "root"
checkmod "$d2/peas" 704 "$USER"
chmod "u+w" "$d1/.ident=$SOMEGROUP/peas"
checkmod "$d1/.ident=$SOMEGROUP/peas" 224 "root"
checkmod "$d1/peas" 724 "$USER"
echo "OK."

echo -n "Enforcing permissions (using user $USER): "
if [ -r ~/"wfs/auth/user_keys/$USER" ] ; then

echo baz > "$d1/fish"
chmod u+r "$d1/.ident=$USER/fish"
check1 "$d3/fish" "baz"
echo baz > "$d1/wish"
chmod u-r "$d1/.ident=$USER/wish"
chmod a-r "$d1/wish"
checknotreadable "$d3/wish"
echo pine > "$d3/yish"
chmod u-w "$d1/.ident=$USER/yish"
chmod a-w "$d1/yish"
checknotwritable "$d3/yish"
#echo '#!/bin/bash' > "$d1/fish" !xavid: bash doesn't like running out of WFS for some reason, but an empty file executes fine
touch "$d1/dish"
chmod u+rwx "$d3/.ident=$USER/dish"
"$d3/dish"
touch "$d1/xish"
chmod u-x "$d3/.ident=$USER/xish"
checkfails "$d3/xish"

echo dish > "$d1/umber"
chown "$SOMEONE" "$d1/umber"
checkfails chmod a+w "$d3/umber"
chown "$USER" "$d1/umber"
chmod a+w "$d3/umber"

# Test things that would be in the cache
echo wombat > "$d3/snail"
check1 "$d3/snail" "wombat"
chmod ua-r "$d3/.ident=$USER/snail"
checknotreadable "$d3/snail"
check1 "$d2/snail" "wombat"

echo wombat > "$d3/trial"
check1 "$d3/trial" "wombat"
chmod ua-r "$d2/.ident=$USER/trial"
checknotreadable "$d3/trial"
check1 "$d2/trial" "wombat"

echo "OK."
else
    echo "no private key found in ~/wfs/auth/user_keys/$USER."
fi

echo "fs-test-10: Passed all tests."
