#!/usr/bin/perl -w
#
# Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
#                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
#     		     Massachusetts Institute of Technology,
# 		     New York University
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

# harness to run file system tests
# cd wfs/test ; ./run-fs-tests.pl

use wfs;

my $NUM_TESTS = 10;
my @do_run = ();
my $debug = 0;
my $secure;

while( $#ARGV > -1 and $ARGV[0] =~ /^-/ ) {
    if ($ARGV[0] eq '-d') {
	$debug++;
    } elsif ($ARGV[0] eq '--secure') {
	$secure='-DWFS_SECURE';
	if( $ARGV[1] !~ /^-/ ) {
	    $secure = $ARGV[1];
	    shift(@ARGV);
	}
    } else {
	die "Unknown arg $ARGV[0]!";
    } 
    shift(@ARGV);
}

# see which tests are set
if( $#ARGV > -1 ) {
    foreach my $t (@ARGV) {
	if( $t <= $NUM_TESTS && $t > 0 ) {
	    $do_run[$t-1] = 1;
	}
    }
} else {
    # turn on all tests
    for( my $i = 0; $i < $NUM_TESTS; $i++ ) {
	$do_run[$i] = 1;
    }
}

my %hosts = ();
my %vars = ();
my %params = ();
my %state = ();

my $localhost_file = "/tmp/localhost.$$.txt";
my $exp_file = "/tmp/exp.$$.txt";

use sigtrap 'handler' => \&cleanup, 'HUP', 'INT', 'ABRT', 'QUIT', 'TERM';

open( H, ">$localhost_file" ) or die( "Couldn't write to $localhost_file" );
print H <<'EOF';
host1 127.0.0.1 27000 1 ${user} ../ /tmp/ ${host_tmp_path}/key_dir
host2 127.0.0.1 27005 2 ${user} ../ /tmp/ ${host_tmp_path}/key_dir
host3 127.0.0.1 27010 3 ${user} ../ /tmp/ ${host_tmp_path}/key_dir
host4 127.0.0.1 27015 4 ${user} ../ /tmp/ ${host_tmp_path}/key_dir
host5 127.0.0.1 27020 5 ${user} ../ /tmp/ ${host_tmp_path}/key_dir
EOF
close(H);

my $user = `whoami`;

open( E, ">$exp_file" ) or die( "Couldn't write to $exp_file" );
print E <<EOF;
var user $user
hostfile $localhost_file
server host1 host2
client host1 host2
cserver host1 host2 host3 host4 host5
command_host host1
command_output stdout
local_output /tmp/wfs.out
debug_level $debug
keep_data 0
paxos_delay_max 2
readahead 0
prefetch-threads 0
EOF
close(E);

parse_exp( $exp_file, \%vars, \%hosts, \%params );

# +xavid
if (defined $secure) {
    my @sec = ($secure);
    $params{'secure'} = \@sec;
    $secure = "-DWFS_SECURE";
} else {
    $secure = "";
}

{;} # stupid emacs

if( $do_run[0] ) {
    run_test( "fs-test-1.pl", './fs-test-1.pl ${clientdir0}' );
}

if( $do_run[1] ) {
    run_test( "fs-test-2.pl", './fs-test-2.pl ${clientdir0} ${clientdir1}' );
}

if( $do_run[2] ) {
    system("cc $secure -o fs-test-3 fs-test-3.c");
    run_test( "fs-test-3", './fs-test-3 ${clientdir0} ${clientdir1}' );
}

if( $do_run[3] ) {
    system("cc $secure -o fs-test-4 fs-test-4.c");
    run_test( "fs-test-4", './fs-test-4 ${clientdir0}/xx' );
}

if( $do_run[4] ) {
    system("cc $secure -o fs-test-5 fs-test-5.c");
    run_test( "fs-test-5", './fs-test-5 ${clientdir0}/fff ${clientdir1}/fff' );
}

if( $do_run[5] ) {
    system("cc $secure -o fs-test-6 fs-test-6.c -lpthread");
    push @{$params{"server"}}, "host3", "host4", "host5";
    push @{$params{"client"}}, "host3", "host4", "host5";
    run_test("fs-test-6", './fs-test-6 ${clientdir0}/.hotspot/fff ${clientdir1}/.hotspot/fff ${clientdir2}/.hotspot/fff ${clientdir3}/.hotspot/fff ${clientdir4}/.hotspot/fff');

    # clean up special stuff
    my @empty = ();
    for( my $i = 0; $i < 3; $i++ ) {
	pop @{$params{"server"}};
	pop @{$params{"client"}};
    }
}

if( $do_run[6] ) {
    push @{$params{"dont_combine"}}, "1";
    system("cc $secure -o fs-test-7 fs-test-7.c");
    run_test( "fs-test-7", './fs-test-7 ${clientdir0} ${clientdir1} ${serverpid0} ${serverpid1}' );
    pop @{$params{"dont_combine"}};
}

if( $do_run[7] ) {
    push @{$params{"ping_period"}}, "2";
    push @{$params{"maint_timer"}}, "1";
    push @{$params{"ping_timeout"}}, "3";
    push @{$params{"dont_combine"}}, "1";
    system("cc $secure -o fs-test-8 fs-test-8.c");
    run_test( "fs-test-8", './fs-test-8 ${clientdir0} ${clientdir1} ${serverpid0} ${serverpid1} ${:host1:_port} ${:host2:_port}' );
    pop @{$params{"ping_period"}};
    pop @{$params{"maint_timer"}};
    pop @{$params{"ping_timeout"}};
    pop @{$params{"dont_combine"}};
 }

if( $do_run[8] ) {
    push @{$params{"ping_period"}}, "2";
    push @{$params{"maint_timer"}}, "1";
    push @{$params{"ping_timeout"}}, "3";
    push @{$params{"dont_combine"}}, "1";
    run_test( "fs-test-9.sh", './fs-test-9.sh ${clientdir0} ${clientdir1} ${serverpid0} ${serverpid1} ${:host1:_port} ${:host2:_port}' );
    pop @{$params{"ping_period"}};
    pop @{$params{"maint_timer"}};
    pop @{$params{"ping_timeout"}};
    pop @{$params{"dont_combine"}};
 }

if( $do_run[9] ) {
    push @{$params{"client"}}, "host3";
    run_test( "fs-test-10.sh", './fs-test-10.sh ${clientdir0} ${clientdir1} ${clientdir2} ${serverpid0} ${serverpid1} ${:host1:_port} ${:host2:_port}' );
    pop @{$params{"client"}};
 }

&log_cleanup();

exit(0);


################################################

sub run_test {

    my $name = shift;
    my $command = shift;

    print "######### Starting test $name #############\n";
    start_wfs( \%vars, \%hosts, \%params, \%state );
    my @e = ($command);
    $params{"command"} = \@e;
    my $st = run_commands( \%vars, \%hosts, \%params, \%state );
    stop_wfs( \%vars, \%hosts, \%params, \%state );
    %state = ();
    if( $st == 0 ) {
	log_cleanup();
	die "$name failed";
    }
    # make sure old sockets are closed:
    sleep(2);

}

sub cleanup() {
    print "run-fs-tests.pl cleaning up on exit ...\n";
    stop_wfs( \%vars, \%hosts, \%params, \%state );
    log_cleanup();
    exit(-1);
}

sub log_cleanup() {
    unlink( $localhost_file );
    unlink( $exp_file );
}

################################################
