# Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
#                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
#     		     Massachusetts Institute of Technology,
# 		     New York University
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
use strict;
use IO::Socket;
use File::Spec;

use Config; 
my $threaded = 0;
if( $Config{usethreads} ) { 
    # We have threads 
    require threads;
    $threaded = 1;
}

my $THREAD_LIMIT = 100;

############################################################################
# Configuration parsing

sub sub_vars {
    my $vars = shift; # hash ref
    my $param = shift;

    my $res = $param;

    foreach my $k (keys(%{$vars})) {
	$res =~ s/\$\{$k\}/$vars->{$k}/g;
    }

    return $res;

}

sub parse_comments_and_vars {

    my $line = shift; # string
    my $vars = shift; # hash ref

    # skip newlines and comments
    if( $line =~ /^\s*$/ or $line =~ /^\s*\#/ ) {
	return 1;
    } elsif( $line =~ /^\s*var\s+(\S+)\s+(.*)$/ ) {
	$vars->{$1} = sub_vars($vars, $2);
	return 1;
    }

    return 0;

}

sub parse_vars {

    my $file = shift; # string
    my $vars = shift; # hash ref

    open( FILE, "<$file" ) or die( "Couldn't open vars $file\n" );
    my @var_lines = <FILE>;
    close(FILE);

    foreach my $line (@var_lines) {
	parse_comments_and_vars( $line, $vars );
    }

}

my @host_fields = qw( name ip port site_id user wfs_path tmp_path key_dir );
sub parse_hosts {

    my $file = shift; # string
    my $vars = shift; # hash ref
    my $hosts = shift; # hash ref

    open( FILE, "<$file" ) or die( "Couldn't open hosts $file\n" );
    my @host_lines = <FILE>;
    close(FILE);

    foreach my $line (@host_lines) {

	if( parse_comments_and_vars( $line, $vars ) ) {
	    next;
	}

	my @fields = split( /\s+/, $line );
	if( $#fields != $#host_fields ) {
	    die( "Bad number of fields in host line: $line\n" );
	}

	my %vals;
	$hosts->{$fields[0]} = \%vals;
	push( @{$hosts->{"ALLHOSTS"}}, $fields[0] );
	for( my $i = 0; $i <= $#fields; $i++ ) {
	    my $val = sub_vars( $vars, $fields[$i] );
	    $hosts->{$fields[0]}->{$host_fields[$i]} = $val;
	}

	# resolve ip if necessary
	$hosts->{$fields[0]}->{"orig_ip"} = $hosts->{$fields[0]}->{"ip"};
	if( !($hosts->{$fields[0]}->{"ip"} =~ /\d+\.\d+\.\d+\.\d+/) ) {
	    my @x_host = gethostbyname($hosts->{$fields[0]}->{"ip"});
	    $hosts->{$fields[0]}->{"ip"} = inet_ntoa( $x_host[4] );
	}

    }

}

sub sub_per_host_vars {
    my $host = shift;
    my $hosts = shift; # hash ref
    my $host_cmd_num = shift;
    my $param = shift;

    my $res = $param;

    foreach my $k (@host_fields, "orig_ip") {
	$res =~ s/\$\{host_$k\}/$hosts->{$host}->{$k}/g;
	$res =~ s/\$\{:(\w+):_$k\}/$hosts->{$1}->{$k}/g;
    }
    $res =~ s/\$\{host_clientdir\}/$hosts->{$host}->{tmp_path}\/fstest-$hosts->{$host}->{port}/g;
    $res =~ s/\$\{host_command_num\}/$host_cmd_num/g;

    return $res;

}

my @exp_params = qw( server client cserver hostfile command command_host command_output local_output debug_level secure keep_data ping_period ping_timeout maint_timer  dont_combine no_permissions paxos_delay_max prefetch-threads read-ahead);
sub parse_exp {

    my $file = shift; # string
    my $vars = shift; # hash ref
    my $hosts = shift; # hash ref 
    my $params = shift; # hash ref

    open( FILE, "<$file" ) or die( "Couldn't open exp $file\n" );
    my @exp_lines = <FILE>;
    close(FILE);

    foreach my $p (@exp_params) {
	my @vals;
	$params->{$p} = \@vals;
    }

    # set special vars
    my $whoami = `whoami`;
    chomp $whoami;
    $vars->{"WHOAMI"} = $whoami;

    foreach my $line (@exp_lines) {

	if( parse_comments_and_vars( $line, $vars ) ) {
	    next;
	}

	my @vals = split( /\s+/, $line );
	my $name = shift(@vals);

	if( $name eq "command" ) {
	    push( @{$params->{$name}}, join(" ", @vals) );
	} else {
	    foreach my $v (@vals) {
		push( @{$params->{$name}}, $v );
	    }
	}

    }

    # parse hosts
    my @all_hosts = ();
    $hosts->{"ALLHOSTS"} = \@all_hosts;
    if( defined $params->{"hostfile"} ) {
	foreach my $h (@{$params->{"hostfile"}}) {
	    parse_hosts( $h, $vars, $hosts );
	}
    }
    
    foreach my $n qw( server client cserver command_host ) {
	if( $#{$params->{$n}} >= 0 && (@{$params->{$n}})[0] eq '${ALLHOSTS}') {
	    pop @{$params->{$n}};
	    foreach my $h (@all_hosts) {
		push( @{$params->{$n}}, $h );
	    }
	}
    }

    # number of hosts
    my @h = keys(%{$hosts});
    $vars->{"NUM_HOSTS"} = ($#h+1); 

    # verify hosts
    foreach my $h (@{$params->{'server'}}, 
		   @{$params->{'client'}}, 
		   @{$params->{'cserver'}},
		   @{$params->{'command_host'}}) {

	if( !defined $hosts->{$h} ) {
	    die( "Cannot find host $h" );
	}
    }

}

############################################################################
# Start-up

my $hostname = `hostname`;
my $ifconfig = `/sbin/ifconfig`;
chomp($hostname);
sub is_localhost {

    my $host = shift; # string
    my $hosts = shift; # hash ref

    if( $hosts->{$host}->{"ip"} eq "127.0.0.1" or
	(gethostbyname($hosts->{$host}->{"ip"}) eq 
	 gethostbyname($hostname)) or 
	($ifconfig =~ /\D$hosts->{$host}->{"ip"}\D/) ) {
	return 1;
    } else {
	return 0;
    }

}

sub blockexec {

    my $host = shift; # string
    my $hosts = shift; # hash ref
    my $log = shift; # string
    my @args = @_;
    my $pid;

    # if it's the localhost, just do a normal system exec
    if( is_localhost( $host, $hosts ) ) {
	my $logging = "";
	if( $log ne "stdout" ) {
	    $logging = " > $log 2>&1";
	}

	return system( join( " ", @args ) . $logging );
    } else {

	if( $log eq "stdout" ) {
	    $log = "/dev/null";
	}

	# otherwise we have to ssh in and start a background process
	my $user = $hosts->{$host}->{"user"};
	my $ip = $hosts->{$host}->{"ip"};
	return system( "ssh $user\@$ip \'@args > $log 2>&1\'" );
    }

}

sub exec_many {

    my $host = shift; # array ref
    my $hosts = shift; # hash ref
    my $log = shift; # array ref
    my $args = shift; # array ref of array refs
    my $fork = shift; # bool
    my @int_pids = ();
    my @out_pids = ();

    my @fork_hosts = @{$host};
    my @logs = @{$log};
    my @args1 = @{$args};
    my %thr_map = ();
    for( my $i = 0; $i <= $#fork_hosts; $i++ ) {

	my $h = $fork_hosts[$i];
	my $l = $logs[$i];
	my @args2 = @{$args1[$i]};

	if( $args2[0] eq "" and $#args2 == 0 ) {
	    push( @int_pids, -1 );
	    next;
	}
	
	my $pid;
	my $func = \&blockexec;
	if( $fork ) {
	    $func = \&forkexec;
	}
	if( $threaded ) {
	    $pid = threads->new( $func, $h, $hosts, $l, @args2 );
	    $thr_map{$pid->tid()} = $i;
	} else {
	    $pid = &$func( $h, $hosts, $l, @args2 );
	}
	push( @int_pids, $pid );

	if( $threaded ) {
	    my @t = threads->list;
	    while( $#t >= $THREAD_LIMIT ) {
		my @joinable;
		if( $threads::VERSION >= 1.67 ) {
		    @joinable = threads->list(0);
		} else {
		    @joinable = threads->list;
		}
		foreach my $thr (@joinable) {
		    if( $thr->tid && !threads::equal($thr, threads->self) ) {
			$pid = $thr->join();
			$int_pids[$thr_map{$thr->tid()}] = -1;
			$out_pids[$thr_map{$thr->tid()}] = $pid;
		    }
		}
		@t = threads->list;
		sleep 1;
	    }
	}

    }

    foreach my $thr (@int_pids) {
	my $pid = $thr;
	if( $threaded and $pid != -1 ) {
	    $pid = $thr->join();
	    $out_pids[$thr_map{$thr->tid()}] = $pid;
	} else {
	    push @out_pids, $pid;
	}
    }

    return @out_pids;

}

sub forkexec {

    my $host = shift; # string
    my $hosts = shift; # hash ref
    my $log = shift; # string
    my @args = @_;
    my $pid;

    # if it's the localhost, just do a normal fork
    if( is_localhost( $host, $hosts ) ) {

	my $logging = "";
	if( $log ne "stdout" ) {
	    $logging = "> $log 2>&1";
	}
	
	$pid = `@args $logging & echo \$!`;
	# just the last line is the pid
	my @lines = split( /\n/, $pid );
	# find the last one that could be a pid
	foreach my $l (@lines) {
	    if( $l =~ /^(\d+)$/ ) {
		$pid = $1;
	    }
	}
	return $pid;

    } else {

	if( $log eq "stdout" ) {
	    $log = "/dev/null";
	}

	# otherwise we have to ssh in and start a background process
	my $user = $hosts->{$host}->{"user"};
	my $ip = $hosts->{$host}->{"ip"};
	$pid = `ssh $user\@$ip \'@args > $log 2>&1 & echo \$!\'`;
	# just the last line is the pid
	my @lines = split( /\n/, $pid );
	# find the last one that could be a pid
	foreach my $l (@lines) {
	    if( $l =~ /^(\d+)$/ ) {
		$pid = $1;
	    }
	}
	return $pid;

    }

}

sub is_linux_many {

    my $host = shift; # array ref
    my $hosts = shift; # hash ref
    my $state = shift; # hash ref

    my @int_pids = ();

    my @fork_hosts = @{$host};
    foreach my $h (@fork_hosts) {

	if( $threaded ) {
	    my $thr = threads->new( \&is_linux, $h, $hosts, $state );
	    push( @int_pids, $thr );
	} else {
	    is_linux( $h, $hosts, $state );
	}

    }

    if( $threaded ) {
	for( my $i = 0; $i <= $#fork_hosts; $i++ ) {
	    my $h = $fork_hosts[$i];
	    my $thr = $int_pids[$i];
	    my $key = $h . "_linux";
	    $state->{$key} = $thr->join();
	}
    }

}

sub is_linux {

    my $host = shift; # string
    my $hosts = shift; # hash ref
    my $state = shift; # hash ref
    my $key = $host . "_linux";

    if( !defined $state->{$key} ) {

	my $uname;
	if( is_localhost( $host, $hosts ) ) {
	    $uname = `uname`;
	} else {
	    # otherwise we have to ssh in and start a background process
	    my $user = $hosts->{$host}->{"user"};
	    my $ip = $hosts->{$host}->{"ip"};
	    $uname = `ssh $user\@$ip \'uname'`;
	}

	if( $uname =~ /linux/i ) {
	    $state->{$key} = 1;
	} else {
	    $state->{$key} = 0;
	}
    }

    return $state->{$key};

}

my @state_fields = qw( cserver_pids server_pids client_pids server_dirs cserver_dirs command_pids client_dirs client_caches tmp_logs );
my $CSERV_MOD = 1;
my $CLIENTPORT_MOD = 2;

sub get_client_precommand {

    my $h = shift; # string
    my $hosts = shift; # hash ref 
    my $state = shift; # hash ref 

    my $port = $hosts->{$h}->{"port"};
    my $tmp_path = $hosts->{$h}->{"tmp_path"};
    
    my $d = "$tmp_path/fstest-$port";
    my $cache = "$tmp_path/fstest-cache-$port";
    my $umount = "umount";
    
    if( is_linux( $h, $hosts, $state ) ) {
	$umount = "fusermount -u";
    }
    
    return ( "$umount $d; ", "rm -rf $d; ", "mkdir $d; ",
	     "rm -rf $cache; ", "mkdir $cache; " );

}

sub get_client_args {

    my $h = shift; # string
    my $hosts = shift; # hash ref 

    my $port = $hosts->{$h}->{"port"};
    my $tmp_path = $hosts->{$h}->{"tmp_path"};
    
    my $d = "$tmp_path/fstest-$port";
    my $cache = "$tmp_path/fstest-cache-$port";
    my $client_port = $port + $CLIENTPORT_MOD;

    return (
	"--mountpoint", "$d",
	"--cache", "$cache",
	"--client-port", "$client_port",
	);
    
}

sub key_dir_init {

    my $h = shift; # string
    my $hosts = shift; # hash ref
    my $secure = shift; # string
    my $key_dir_cmd = shift; # string

    if( $secure ) {
	my $remote = "";
	if( !is_localhost( $h, $hosts ) ) {
	    my $user = $hosts->{$h}->{"user"};
	    my $ip = $hosts->{$h}->{"ip"};
	    $remote = "$user@$ip:";
	}
	
	system( "rsync -r $secure/ $remote$key_dir_cmd" );
    }

}

sub start_wfs {

    my $vars = shift; # hash ref
    my $hosts = shift; # hash ref 
    my $params = shift; # hash ref
    my $state = shift; # hash ref
    
    foreach my $f (@state_fields) {
	my @vals;
	$state->{$f} = \@vals;
    }

    my @lo = @{$params->{"local_output"}};
    my $local_output = "/dev/null";
    if( defined $lo[0] ) {
	$local_output = $lo[0];
    }

    my @d = @{$params->{"debug_level"}};
    my $debug_level = 0;
    if( $#d >= 0 ) {
	$debug_level = $d[$#d];
    };
    my @debug = ();
    while( $debug_level > 0 ) {
	push( @debug, "--debug" );
	$debug_level--;
    }
    
    # +xavid
    my $secure = "";
    my @sec = @{$params->{"secure"}};
    if ($#sec >= 0) {
	if ($sec[0]) {
	    $secure=$sec[0];
	} else {
	    $secure='${host_wfs_path}/auth';
	}
	print "Using secure connections with keydir $secure\n";
    }

    # +irene
    my $read_ahead = 0;
    my @read = @{$params->{"readahead"}}; 
    if ($#read >= 0) {
	$read_ahead = $read[0];
    }
    my $prefetch_threads = 0;
    my @prefetcht = @{$params->{"prefetch-threads"}};
    if( $#prefetcht >= 0 ) {
	$prefetch_threads = $prefetcht[0];
    }    
    
    my $keep_data = 0;
    my @k = @{$params->{"keep_data"}};
    if( $#k >= 0 ) {
	$keep_data = $k[0];
    }

    my $no_permissions = 0;
    my @np = @{$params->{"no_permissions"}};
    if( $#np >= 0 ) {
	$no_permissions = $np[0];
    }

    my $dont_combine = 0;
    my @dc = @{$params->{"dont_combine"}};
    if( $#dc >= 0 ) {
	$dont_combine = $dc[0];
    }

    my $ping_period = 0;
    my @pp = @{$params->{"ping_period"}};
    if( $#pp >= 0 ) {
	$ping_period = $pp[0];
    }

    my $ping_timeout = 0;
    my @pt = @{$params->{"ping_timeout"}};
    if( $#pt >= 0 ) {
	$ping_timeout = $pt[0];
    }

    my $paxos_delay_max = 0;
    my @px = @{$params->{"paxos_delay_max"}};
    if( $#px >= 0 ) {
	$paxos_delay_max = $px[0];
    }

    my $maint_timer = 0;
    my @mt = @{$params->{"maint_timer"}};
    if( $#mt >= 0 ) {
	$maint_timer = $mt[0];
    }

    # start cservers
    my @cservers = @{$params->{"cserver"}};
    my $cserv_ip;
    my $cserv_port;
    foreach my $h (@cservers) {
	my $ip = $hosts->{$h}->{"ip"};
	my $port = $hosts->{$h}->{"port"} + $CSERV_MOD;
	my $path = File::Spec->rel2abs($hosts->{$h}->{"wfs_path"});
	my $tmp_path = $hosts->{$h}->{"tmp_path"};
	my $key_dir = $hosts->{$h}->{"key_dir"} . "_$port";
	my $d = "$tmp_path/cserver.$port.logs";
	if( !defined $cserv_ip ) {
	    $cserv_ip = $ip;
	    $cserv_port = $port;
	}
	my $cserver_log = tmp_log($local_output, 0, $h, $hosts);
	if( $local_output ne "/dev/null" and $local_output ne "stdout" ) {
	    $cserver_log .= ".cserver.$port";
	}

	my @extra_args = ();
	if( $ping_timeout > 0 ) {
	    push @extra_args, "--timeout", "$ping_timeout";
	}
	if( $paxos_delay_max > 0 ) {
	    push @extra_args, "--paxos-delay-max", "$paxos_delay_max";
	}
	my $key_dir_cmd = "";
	if ($secure) {
	    $key_dir_cmd = sub_per_host_vars($h,$hosts,0,$key_dir);
	    push @extra_args, "--secure","--keydir", $key_dir_cmd;
	}

	blockexec( $h, $hosts, $local_output, "mkdir -p $d $key_dir_cmd" );
	key_dir_init( $h, $hosts, $secure, $key_dir_cmd );
	my $pid = forkexec( $h, $hosts, $cserver_log, 
			    "cd $tmp_path; ulimit -c unlimited; cd $tmp_path; ",
			    "nohup",
#			    "valgrind", "--leak-check=yes","--suppressions=$path/test/libcrypto.supp",
			    "$path/cserver/wfs_rsm_cserver",
			    "--master", "$cserv_ip:$cserv_port", 
			    "--this-host", "$ip:$port", 
			    "--log-dir", "$d",
			    "--dbg-dir", "$d",
			    @extra_args,
			    @debug );
	push( @{$state->{"cserver_pids"}}, $pid );
	push( @{$state->{"cserver_dirs"}}, "$d $key_dir_cmd" );
    }

    # wait for the cservers to settle on a view before start anyone else
    my $last_view_size = 0;
    print "Waiting for cservers to settle...\n";
    while( $last_view_size < ($#cservers+1) ) {
	my $h = $cservers[0];
	my $ip = $hosts->{$h}->{"ip"};
	my $port = $hosts->{$h}->{"port"} + $CSERV_MOD;
	my $tmp_path = $hosts->{$h}->{"tmp_path"};
	my $d = "$tmp_path/cserver.$port.logs";
	my $view_out = "";
	my $view_cmd = "grep view $d/paxos*log | tail -1";
	if( is_localhost( $h, $hosts ) ) {
	    $view_out = `$view_cmd`;
	} else {
	    # otherwise we have to ssh in and start a background process
	    my $user = $hosts->{$h}->{"user"};
	    my $ip = $hosts->{$h}->{"ip"};
	    my $stderr = "/tmp/stderr." . rand 0xffffffff;
	    $view_out = `ssh $user\@$ip \"$view_cmd\" 2> $stderr`;
	    system( "rm -f $stderr" );
	}
	my @v = split( /\s+/, $view_out );
	$last_view_size = $v[2];
	if( $last_view_size < ($#cservers+1) ) {
	    sleep(1);
	}
    }

    # start servers
    my %started_servers;
    my @servers = @{$params->{"server"}};
    my @clients = @{$params->{"client"}};
    my $i = 0;

    my @fm_hosts = ();
    my @fm_logs = ();
    my @fm_args = ();
    my @s_dirs = ();
    my @sk_dirs = ();

    foreach my $h (@servers) {

	# am I also a client?  If so, merge
	my $client_also = 0;
	if( !$dont_combine ) {
	    foreach my $c (@clients) {
		if( $c eq $h ) {
		    $client_also = 1;
		}
	    }
	}

	my $ip = $hosts->{$h}->{"ip"};
	my $port = $hosts->{$h}->{"port"};
	my $site_id = $hosts->{$h}->{"site_id"};
	my $path = $hosts->{$h}->{"wfs_path"};
	my $tmp_path = $hosts->{$h}->{"tmp_path"};
	my $key_dir = $hosts->{$h}->{"key_dir"}  . "_$port";
        my $d = "$tmp_path/dir$port";
	push( @s_dirs, "$d" );

	my $rmdir = "echo";
	if( !$keep_data ) {
	    $rmdir = "rm -rf $d";
	}
	
	my @args;

	push( @args, "$rmdir; ", "mkdir $d; ", 
	      "echo '1,0000000000000000, ,1,00000001,anyuser=00000007' > $d/0000000000000001.m; ", 
	      "touch $d/0000000000000001; ", 
	      "chmod 666 $d/0000000000000001.m; ",
	      "chmod 777 $d/0000000000000001;",
#	      "ulimit -s 512; ", 
	      "ulimit -c unlimited; cd /tmp/; " 
    );
	my $server_log = tmp_log($local_output, 0, $h, $hosts);
	if( $local_output ne "/dev/null" and $local_output ne "stdout" ) {
	    if( $client_also ) {
		$server_log .= ".$port";
	    } else {
		$server_log .= ".server.$port";
	    }
	}

	if( $client_also ) {
	    push( @args, get_client_precommand( $h, $hosts, $state ) );
	}

	# localhost needs special treatment because it uses exec()
	if( is_localhost($h, $hosts) ) {
	    blockexec( $h, $hosts, $local_output, @args );
	    @args = ();
	}

	my @extra_args = ();
	if( $ping_period > 0 ) {
	    push @extra_args, "--ping-period", "$ping_period";
	}
	if( $maint_timer > 0 ) {
	    push @extra_args, "--maint-timer", "$maint_timer";
	}

	if( $no_permissions ) {
	    push @extra_args, "--no-permissions";
	}
	my $key_dir_cmd = "";
	if ($secure) {
	    $key_dir_cmd = sub_per_host_vars($h,$hosts,0,$key_dir);
	    push @extra_args, "--secure","--keydir", $key_dir_cmd;
	}

	# +irene
	if ($read_ahead) {
	    push @extra_args, "--read-ahead";
	}
	if( $prefetch_threads > 0) {
	    push @extra_args, "--prefetch-threads", "$prefetch_threads";
	}
	
	my $progname = "wfs_server";
	if( $client_also ) {
	    $progname = "wfs";
	    push( @extra_args, get_client_args( $h, $hosts ) );
	}

	push( @sk_dirs, "$key_dir_cmd" );

	key_dir_init( $h, $hosts, $secure, $key_dir_cmd );
	push( @args,
	      "nohup",
#	      "valgrind", "--leak-check=yes","--suppressions=$path/test/libcrypto.supp",
	      "$path/fs/$progname",
	      "--addr", "$ip",
	      "--port", "$port",
	      "--site", "$site_id",
	      "--dir", "$d",
	      "--dbg-dir", "$tmp_path",
	      "--cserv-master", "$cserv_ip:$cserv_port",
	      @extra_args,
	      @debug );

	push( @fm_hosts, $h );
	push( @fm_logs, $server_log );
	push( @fm_args, \@args );

    }

    my @s_pids = exec_many( \@fm_hosts, $hosts, \@fm_logs, \@fm_args, 1 );

    foreach my $h (@servers) {
	my $pid = $s_pids[$i];
	my $d = $s_dirs[$i];
	my $dk = $sk_dirs[$i];
	push( @{$state->{"server_pids"}}, $pid );
	push( @{$state->{"server_dirs"}}, "$d $dk" );
	$started_servers{$h}=1;
	$vars->{"serverpid$i"} = $pid;
	$i++;
    }

    # servers need a second to connect with the node manager, etc.
    print "Waiting for servers to settle...\n";
    sleep(10); # Needed to increase this for the fuse mounts to be reliably up when using --secure; using mountpoint(1) to check for this would be better (but it's not on Mac)

    # start clients

    @fm_hosts = ();
    @fm_logs = ();
    @fm_args = ();
    my @c_dirs = ();
    my @ck_dirs = ();
    my @c_caches = ();

    # first cache the is_linux state
    is_linux_many( \@clients, $hosts, $state );

    my $actual_clients = 0;
    foreach my $h (@clients) {

	my $server_also = 0;
	if( !$dont_combine ) {
	    foreach my $s (@servers) {
		if( $s eq $h ) {
		    $server_also = 1;
		}
	    }
	}
	#print "client $h server_also=$server_also\n";
	my $ip = $hosts->{$h}->{"ip"};
	my $port = $hosts->{$h}->{"port"};
	my $client_port = $port + $CLIENTPORT_MOD;
	my $path = $hosts->{$h}->{"wfs_path"};
	my $tmp_path = $hosts->{$h}->{"tmp_path"};
	my $key_dir = $hosts->{$h}->{"key_dir"} . "_$port";
	my $site_id = $hosts->{$h}->{"site_id"};

        my $d = "$tmp_path/fstest-$port";
	my $cache = "$tmp_path/fstest-cache-$port";
	push( @c_caches, $cache );
	push( @c_dirs, "$d" );

	my @args = get_client_precommand( $h, $hosts, $state );

	if( !$server_also and is_localhost( $h, $hosts ) ) {
	    blockexec( $h, $hosts, $local_output, @args );
	    @args = ();
	}

	my @client_args = get_client_args( $h, $hosts );

	my @extra_args;
	if( !defined($started_servers{$h}) ) {
	    my $remote_ip = $hosts->{$servers[0]}->{"ip"};
	    $port = $hosts->{$servers[0]}->{"port"};
	    push @extra_args, "--addr", "$remote_ip";
	} else {
	    push @extra_args, "--addr", "$ip";
	}

	if( $ping_period > 0 ) {
	    push @extra_args, "--ping-period", "$ping_period";
	}

	if( $no_permissions ) {
	    push @extra_args, "--no-permissions";
	}

	# +irene
	if( $prefetch_threads > 0) {
	    push @extra_args, "--prefetch-threads", "$prefetch_threads";
	}
	
	my $client_log = tmp_log($local_output, 0, $h, $hosts);
	if( $local_output ne "/dev/null" and $local_output ne "stdout" ) {
	    $client_log .= ".client.$client_port";
	}
	my $key_dir_cmd = "";
	if ($secure) {
	    $key_dir_cmd = sub_per_host_vars($h,$hosts,0,$key_dir);
	    push @extra_args, "--secure","--keydir", $key_dir_cmd;
	}

	push( @ck_dirs, "$key_dir_cmd" );
	push( @extra_args, get_client_args( $h, $hosts ) );

	key_dir_init( $h, $hosts, $secure, $key_dir_cmd );
	push( @args,
	      "ulimit -c unlimited; ",
	      "nohup",
#	      "valgrind", "--leak-check=yes", "--suppressions=$path/test/libcrypto.supp",
	      "$path/fs/wfs_fuse_client",
	      "--port", "$port",
	      "--site", "$site_id",
	      "--dbg-dir", "$tmp_path",
	      "--cserv-master", "$cserv_ip:$cserv_port",
	      @extra_args,
	      @debug );

	# only need to really execute these args if we weren't a server as well
	if( $server_also ) {
	    @args = ( "" );
	} else {
	    $actual_clients++;
	}

	push( @fm_hosts, $h );
	push( @fm_logs, $client_log );
	push( @fm_args, \@args );

    }
    my @c_pids = exec_many( \@fm_hosts, $hosts, \@fm_logs, \@fm_args, 1 );

    $i = 0;
    foreach my $h (@clients) {

	my $pid = $c_pids[$i];
	my $d = $c_dirs[$i];
	my $dk = $ck_dirs[$i];
	my $cache = $c_caches[$i];
	push( @{$state->{"client_pids"}}, $pid );
	push( @{$state->{"client_dirs"}}, "$d $dk" );
	push( @{$state->{"client_caches"}}, $cache );
	$vars->{"clientdir$i"} = $d;
	$vars->{"clientpid$i"} = $pid;

	$i++;
    }

    if( $actual_clients > 0 ) {
	print "Waiting for clients to settle...\n";
	sleep(10); # See note by "Waiting for servers to settle"
    }
}

sub stop_wfs {

    my $vars = shift; # hash ref
    my $hosts = shift; # hash ref 
    my $params = shift; # hash ref
    my $state = shift; # hash ref

    # commands killed first
    my @command_hosts = @{$params->{"command_host"}};
    my @command_pids = @{$state->{"command_pids"}};

    my $keep_data = 0;
    my @k = @{$params->{"keep_data"}};
    if( $#k >= 0 ) {
	$keep_data = $k[0];
    }

    my %cmds = ();

    for( my $i = 0; $i <= $#command_pids; $i++ ) {

	my $cmd_host = $command_hosts[$i];
	my $pid = $command_pids[$i];

	if( $pid == -1 ) {
	    next;
	} else {
	    my $command = "kill $pid; sleep 2; ";
	    if( defined $cmds{$cmd_host} ) {
		$cmds{$cmd_host} .= $command;
	    } else {
		$cmds{$cmd_host} = $command;
	    }
	}

    }

    my @pids = (@{$state->{"client_pids"}}, @{$state->{"server_pids"}},
		@{$state->{"cserver_pids"}});
    my @dirs = (@{$state->{"client_dirs"}}, @{$state->{"server_dirs"}},
		@{$state->{"cserver_dirs"}});
    my @caches = (@{$state->{"client_caches"}});
    my @on_hosts = (@{$params->{"client"}}, @{$params->{"server"}}, 
		    @{$params->{"cserver"}});
    
    for( my $i = 0; $i <= $#pids; $i++ ) {

	# assume user command has finished
#	my $command = "kill -INT $pids[$i]; ";
#	my $command = "kill -ABRT $pids[$i]; ";
	my $command = "kill -9 $pids[$i]; ";
#	my $command = "gcore $pids[$i] ; kill -9 $pids[$i]; ";

	if( $pids[$i] == -1 ) {
	    $command = "";
	}

	if( defined $dirs[$i] && !$keep_data ) {
	    $command = "rm -rf $dirs[$i]; " . $command;
	}

	my $umount = "umount";
	if( is_linux( $on_hosts[$i], $hosts, $state ) ) {
	    $umount = "fusermount -u";
	}

	if( defined $caches[$i] ) {
	    $command = "rm -rf $caches[$i]; $umount $dirs[$i]; " . $command;
	}

	if( defined $cmds{$on_hosts[$i]} ) {
	    $cmds{$on_hosts[$i]} .= $command;
	} else {
	    $cmds{$on_hosts[$i]} = $command;
	}

    }

    # clean up tmp logs
    my @logs = (@{$state->{"tmp_logs"}});
    foreach my $log (@logs) {
	my @parts = split( /\:/, $log );
	if( $#parts == 1 ) {
	    my $c = "rm -f $parts[1]; ";
	    if( defined $cmds{$parts[0]} ) {
		$cmds{$parts[0]} .= $c;
	    } else {
		$cmds{$parts[0]} = $c;
	    }
	}
    }

    my @fm_hosts = ();
    my @fm_args = ();
    my @fm_logs = ();
    foreach my $h (keys(%cmds)) {

	my @x = ();
	push( @x, $cmds{$h} );

	if( is_localhost($h, $hosts) ) {
	    blockexec( $h, $hosts, "/dev/null", @x );
	} else {
	    push( @fm_hosts, $h );
	    push( @fm_args, \@x );
	    push( @fm_logs, "/dev/null" );
	}
    }

    exec_many( \@fm_hosts, $hosts, \@fm_logs, \@fm_args, 0 );

}

############################################################################
# Command execution

sub tmp_log {

    my $log = shift; # string
    my $need_suffix = shift; # bool
    my $host = shift; # string
    my $hosts = shift; # hash ref

    # just the tmp_path + file name
    my @parts = split( /\//, $log );
    my $tmp_path = $hosts->{$host}->{"tmp_path"};
    my $port = $hosts->{$host}->{"port"};

    my $suffix = "";
    if( $need_suffix ) {
	$suffix = ".$port";
	if( is_localhost( $host, $hosts ) ) {
	    $suffix .= ".tmp";
	}
    }

    return $tmp_path . "/" . $parts[$#parts] . $suffix;

}

sub append_remote_logs {

    my $log = shift; # string
    my $host = shift; # array ref
    my $remote_log = shift; # array ref
    my $hosts = shift; # hash ref

    my @int_pids = ();

    my @fork_hosts = @{$host};
    my @remote_logs = @{$remote_log};
    my @tmp_files = ();
    for( my $i = 0; $i <= $#fork_hosts; $i++ ) {

	my $h = $fork_hosts[$i];
	my $rl = $remote_logs[$i];

	my $tmp_file = "/tmp/wfstmp." . rand(0xffffffff);
	push( @tmp_files, $tmp_file );

	if( $threaded ) {
	    my $thr = threads->new( \&get_remote_log, $h, $rl, 
				    $tmp_file, $hosts );
	    push( @int_pids, $thr );
	} else {
	    get_remote_log( $h, $rl, $tmp_file, $hosts );
	}

# 	while( $threaded and scalar(threads->list()) >= $THREAD_LIMIT ) {
# 	    my $j = 0;
# 	    while( $int_pids[$j] == -1 ) {
# 		$j++;
# 	    }
# 	    $int_pids[$j]->join();
# 	    $int_pids[$j] = -1;
# 	}

    }

    my @fm_logs = ();
    my @fm_args = ();
    for( my $i = 0; $i <= $#tmp_files; $i++ ) {
	my $tmp_file = $tmp_files[$i];
	my $rl = $remote_logs[$i];
	if( $threaded ) {
	    my $thr = $int_pids[$i];
	    if( $thr != -1 ) {
		$thr->join();
	    }
	}
	system( "cat $tmp_file >> $log; rm -f $tmp_file" );
    }

}

sub get_remote_log {

    my $host = shift; # string
    my $remote_log = shift; # string
    my $tmp_file = shift; # string
    my $hosts = shift; # hash ref

    # if it's the localhost, just do a normal system exec
    if( is_localhost( $host, $hosts ) ) {
	return system( "cp $remote_log $tmp_file" );
    } else {
	# otherwise we have to ssh in and start a background process
	my $user = $hosts->{$host}->{"user"};
	my $ip = $hosts->{$host}->{"ip"};
	my $s = system( "scp $user\@$ip:$remote_log $tmp_file" );
	if( $s != 0 ) {
	    return $s;
	}
    }

}

sub wait_for_commands {

    my $hosts = shift; # hash ref
    my $params = shift; # hash ref
    my $state = shift; # hash ref
    my $ch = shift; # hash ref
    my $cp = shift; # array ref

    my @command_hosts = @{$ch};
    my @command_pids = @{$cp};

    my $done = 0;

    do {

	$done = 0;

	# pre-count the done dudes
	for( my $i = 0; $i <= $#command_hosts; $i++ ) {
	    my $pid = $command_pids[$i];
	    if( $pid == -1 ) {
		my $host = $command_hosts[$i];
		$done++;
		next;
	    }
	}

	for( my $i = 0; $i <= $#command_hosts; $i++ ) {
	
	    my $host = $command_hosts[$i];
	    my $pid = $command_pids[$i];
	    
	    if( $pid == -1 ) {
		next;
	    }
	    
	    my $proc_done = 0;

	    # if it's the localhost, just do a normal system exec
	    my $ps_cmd = "ps -p $pid";
	    my $ps_out;
	    if( is_localhost( $host, $hosts ) ) {
		$ps_out = `$ps_cmd`;
	    } else {
		# otherwise we have to ssh in and start a background process
		my $user = $hosts->{$host}->{"user"};
		my $ip = $hosts->{$host}->{"ip"};
		my $stderr = "/tmp/stderr." . rand 0xffffffff;
		$ps_out = `ssh $user\@$ip \"$ps_cmd\" 2> $stderr`;
		# planetlab . . . 
		if( `grep -c proc $stderr` != 0 ) {
		    print "Got proc in stderr, retrying\n";
		    $ps_out = "$pid";
		}
		system( "rm -f $stderr" );
	    }

	    if( !($ps_out =~ /$pid/) ) {
		$proc_done = 1;
	    }

	    if( $proc_done ) {
		# the process isn't running anymore
		$command_pids[$i] = -1;
		@{$state->{"command_pids"}}[$i] = -1;
		$done++;
	    } else {
		sleep(10);
	    }

	}

    } while( $done != ($#command_pids+1) );

}

sub wait_for_commands_many {

    my $hosts = shift; # hash ref
    my $params = shift; # hash ref
    my $state = shift; # hash ref
    my $ch = shift; # hash ref
    my @command_hosts = @{$ch};
    my @command_pids = @{$state->{"command_pids"}};
    
    # split up into THREAD_LIMIT lists and wait for all at once
    my @host_lists = ();
    my @pid_lists = ();
    my $i = 0;
    foreach my $h (@command_hosts) {
	if( !defined $host_lists[$i%$THREAD_LIMIT] ) {
	    my @list1 = ();
	    my @list2 = ();
	    $host_lists[$i%$THREAD_LIMIT] = \@list1;
	    $pid_lists[$i%$THREAD_LIMIT] = \@list2;
	}
	push @{$host_lists[$i%$THREAD_LIMIT]}, $h;
	push @{$pid_lists[$i%$THREAD_LIMIT]}, $command_pids[$i];
	$i++;
    }

    # fork a thread for each
    my @int_pids = ();
    for( my $j = 0; $j <= $#host_lists; $j++ ) {
	if( $threaded ) {
	    my $thr = threads->new( \&wait_for_commands, $hosts, $params, 
				    $state, $host_lists[$j], $pid_lists[$j] );
	    push( @int_pids, $thr );
	} else {
	    wait_for_commands( $hosts, $params, 
			       $state, $host_lists[$j], $pid_lists[$j] );
	}
    }

    if( $threaded ) {
	for( my $j = 0; $j <= $#host_lists; $j++ ) {
	    my $thr = $int_pids[$j];
	    $thr->join();
	}
    }

    my @no_pids;
    $state->{"command_pids"} = \@no_pids;

}

sub run_commands {

    my $vars = shift; # hash ref
    my $hosts = shift; # hash ref 
    my $params = shift; # hash ref
    my $state = shift; # hash ref

    my @command_hosts = @{$params->{"command_host"}};
    my @commands = @{$params->{"command"}};

    # if there's an output file, clear it
    my @command_output = @{$params->{"command_output"}};
    my $log;
    if( $#command_output >= 0 ) {
	$log = $command_output[0];
	truncate( $log, 0 );
    }

    my $i = 0;
    my @last_fm_hosts = ();
    my @last_fm_logs = ();
    foreach my $command (@commands) {
	#print "Running $command...\n";
	my @ch = ();

	my $block = 0;
	if( $command =~ /^(b?)\@([\d\,]+)\s+(.*)$/ ) {
	    if( $1 eq "b" ) {
		$block = 1;
	    }
	    my @h = split( /\,/, $2 );
	    foreach my $d (@h) {
		push( @ch, $command_hosts[$d] );
	    }
	    $command = $3;
	} else {
	    @ch = @command_hosts;
	}

	if( $#ch == 0 ) {
	    $block = 1;
	}

	my @fm_hosts = ();
	my @fm_logs = ();
	my @fm_args = ();

	$vars->{"NUM_COMMANDED"} = ($#ch+1);
	$command = sub_vars( $vars, $command );
	my $cmd_num = 0;

	foreach my $cmd_host (@ch) {

	    my $remote_log = "/dev/null";

	    if( defined $log and $log ne "stdout" ) {
		$remote_log = tmp_log( $log, 1, $cmd_host, $hosts );
		push( @{$state->{"tmp_logs"}}, 
		      "$cmd_host:$remote_log" );
	    } elsif( defined $log and $log eq "stdout" ) {
		$remote_log = $log;
	    }

	    my $host_command = sub_per_host_vars( $cmd_host, $hosts, $cmd_num,
						  $command );
	    $cmd_num++;
	    
	    push( @fm_hosts, $cmd_host );
	    push( @fm_logs, $remote_log );
	    my @x = ();
	    if( $i != 0 ) {
		push( @x, "rm -rf $remote_log; " );
	    }
	    $i++;
	    push( @x, $host_command );
	    push( @fm_args, \@x );

	}

	my @pids = exec_many( \@fm_hosts, $hosts, \@fm_logs, 
			      \@fm_args, !$block );

	if( !$block ) {
	    foreach my $pid (@pids) {
		push( @{$state->{"command_pids"}}, $pid );
	    }
	    &wait_for_commands_many( $hosts, $params, $state, \@ch );
	} else {
	    if( $pids[0] != 0 ) {
		#print "pids[0] != 0!\n";
		return 0;
	    }
	}

	@fm_hosts = ();
	@fm_logs = ();
	@fm_args = ();

	foreach my $cmd_host (@ch) {

	    my $remote_log = "/dev/null";

	    if( defined $log and $log ne "stdout" ) {
		$remote_log = tmp_log( $log, 1, $cmd_host, $hosts );
	    } elsif( defined $log and $log eq "stdout" ) {
		$remote_log = $log;
	    }

	    # copy the output to the local log
	    if( defined $log and $log ne "stdout" ) {
		push( @fm_hosts, $cmd_host );
		push( @fm_logs, $remote_log );
		pop( @{$state->{"tmp_logs"}} );
	    }
	}
	append_remote_logs( $log, \@fm_hosts, \@fm_logs, $hosts );
	@last_fm_hosts = @fm_hosts;
	@last_fm_logs = @fm_logs;
    }

    # last log removal
    my @fm_args;
    my @fm_logs;
    foreach my $log (@last_fm_logs) {
	my @x = ();
	push( @x, "rm -rf $log" );
	push( @fm_args, \@x );
	push( @fm_logs, "/dev/null" );
    }
    exec_many( \@last_fm_hosts, $hosts, \@fm_logs, \@fm_args, 0 );

    return 1;

}

1;
