/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef __WFS_LEASE_MANAGER_H__
#define __WFS_LEASE_MANAGER_H__ 1

#include "wfs_file.h"

#include <assert.h>
#include <ext/hash_map>
#include <pthread.h>
#include <vector>
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;

struct node_id_hash {
  std::size_t operator()( const wfs_node_id &a ) const {
    return hash<uint32_t>()( (uint32_t) (a>>32) ) ^ hash<uint32_t>()( (uint32_t) ((a<<32)>>32) );
  }
};

class wfs_lease_manager {

 public:

  const static int32_t LEASE_BATCH_MARGIN;

  typedef struct lease {
    wfs_file_handle fh;
    wfs_node node;
    time_t expiration; // absolute time, in UTC secs
    int32_t stale_time;    
    bool recently_used;
    lease *next_by_fid;
    lease *prev_by_node;
    lease *next_by_node;
  lease() : expiration(0), stale_time(0), recently_used(true),
      next_by_fid(NULL), prev_by_node(NULL), next_by_node(NULL) {}

  } lease;

 private:

  /**
   * Leases are held in two data structures:
   *   - A map mapping FID -> singly-linked list of leases on different nodes
   *   - A map mapping node ids -> doubly-linked list of leases of 
   *     different files
   */

  pthread_mutex_t _lease_mutex;
  hash_map< wfs_file_id, lease * > _leases_by_fid;
  std::vector<lease *> _new_leases;
  hash_map< wfs_node_id, lease *, node_id_hash > _leases_by_node;
  hash_map< wfs_node_id, lease *, node_id_hash > _last_leases_by_node;
  lease *_next_expiration;
  time_t _last_batch;
  pthread_cond_t _expire_cond;
  bool _in_wait_thread;

  hash_map< wfs_file_id, int > _pending_leases;
  pthread_mutex_t _pending_leases_lock;

  lease *internal_add_lease( wfs_file_handle fh, wfs_node node, 
			     time_t expiration /*absolute*/,
			     int32_t stale_time );
  void internal_remove_node_expiration( lease *xlease );
  void internal_update_expiration_state( lease *xlease,
					 time_t old_expiration );
  void internal_remove_lease( lease *xlease );
  void internal_next_expiration();
  void add_to_node_list( lease *xlease );
  void remove_from_node_list( lease *xlease );

 public:

  typedef struct leaser {
    wfs_node node;
    time_t time_left;
    int32_t stale_time;
  } leaser;

  wfs_lease_manager();
  ~wfs_lease_manager();
  void add_lease( wfs_file_handle fh, wfs_node node, 
		  time_t deadline /*relative*/,
		  int32_t stale_time = 0 );
  bool has_lease( wfs_file_handle fh, wfs_node node );
  void invalidate_lease( wfs_file_handle fh, wfs_node *modifier = NULL );
  void leasing_nodes( wfs_file_handle fh, std::vector<leaser> &nodes );
  void wait_for_next_batch( wfs_node &node, std::vector<wfs_file_handle> &fhs );
  void renew_leases( wfs_node node, std::vector<wfs_file_handle> fh );
  void reg_pending_lease( wfs_file_handle fh );
  bool unreg_pending_lease( wfs_file_handle fh );
  
};

#endif // __WFS_LEASE_MANAGER_H__
