/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_metadata.h"

#include <fcntl.h>
#include <assert.h>

#include "wfs_misc.h"

std::string wfs_metadata::_dir = "";

wfs_metadata::wfs_metadata( const wfs_metadata &md ) {

  _filename = md._filename;
  fid = md.fid;
  version = md.version;
  perm_cues = md.perm_cues;
  permissions = md.permissions;
  num_links = md.num_links;

}

wfs_metadata::wfs_metadata( wfs_file_id id, int32_t vernum ) :
  fid(id), perm_cues(""), num_links(0) {
  if (vernum == 0) {
    char buf[WFS_FID_NUM_HEX_BYTES+5];
    snprintf( buf, WFS_FID_NUM_HEX_BYTES+5, "/%016"PRIx64".m", fid );
    _filename = _dir + std::string( buf, strlen(buf) );
  } else {
    char buf[WFS_FID_NUM_HEX_BYTES+5+1+WFS_INT_NUM_DEC_BYTES];
    snprintf( buf, WFS_FID_NUM_HEX_BYTES+5+1+WFS_INT_NUM_DEC_BYTES,
	      "/%016"PRIx64".m@%d", fid, vernum );
    _filename = _dir + std::string( buf, strlen(buf) );
  }

  read_in_file();

}

wfs_metadata::~wfs_metadata() {
}

void wfs_metadata::delete_file() {
  unlink(_filename.c_str());
}

/**
 * Format of metadata file (comma separated fields:
 *   Version number (4 bytes)
 *   Version node id (8 bytes)
 *   permanent cues + space
 *   number of specified replicas (4 bytes)
 *   for the number of replicas:
 *     ip address + port combinations (ip_int:port_int)
 *   number of ACL entries (4 bytes)
 *   for the number of entries:
 *     name (up to MAX_IDENT_LENGTH chars) and permissions in hex (4 bytes) (name=perm)
 */

int max_md_size = 1024;

bool wfs_metadata::read_in_file() {

  int fd = open( _filename.c_str(), O_RDONLY );

  if( fd < 0 ) {
    return false;
  }

  // read everything in at first:
  char buf[max_md_size];
  int n = read( fd, buf, max_md_size );

  wfs_log(WFS_DBG_2, "Read %d metadata bytes for file %016"PRIx64"\n", n, fid );

  assert( n < max_md_size );
  char *p;
  char *x;

  // read the version
  uint32_t version_num = 0;
  wfs_node_id node_id = 0;

  wfs_log(WFS_DBG_2, "md checkpoint\n" );

  int scanres;

  if( n > 0 ) {
    buf[n] = '\0';
    wfs_log(WFS_DBG_2, "wfs_metadata:read_in_file: %s\n", buf);
    p = strtok_r( buf, ",", &x );
    assert( p != NULL );
    version_num = atoi(p);
    wfs_log(WFS_DBG_2, "version num (%s):%d\n", buf, version_num );
    p = strtok_r( NULL, ",", &x );
    assert( p != NULL );
    wfs_log(WFS_DBG_2, "node id (%s):%s\n", buf, p );
    uint32_t id1;
    uint32_t id2;
    scanres = sscanf( p, "%08x%08x", &id1, &id2 );
    assert (scanres == 2);
    node_id = two_ints_to_node_id( id1, id2 );
    if( wfs_dbg_block(WFS_DBG_2) ) {
      wfs_log(WFS_DBG_2, "id1: %08x\n", id1);
      wfs_log(WFS_DBG_2, "id2: %08x\n", id2);
      wfs_log(WFS_DBG_2, "node id: %016"PRIx64"\n", node_id );
    }
    version = wfs_version( version_num, node_id );
  } else {
    close(fd);
    return false;
  }

  // read perm cues
  p = strtok_r( NULL, ",", &x );
  assert( p != NULL );
  perm_cues = std::string( p, strlen(p)-1 /*-1 for space*/ );

  // get number of hard links to file
  p = strtok_r( NULL, ",", &x );
  assert( p != NULL );
  num_links = atoi(p);

  // Permissions
  permissions.clear();
  // get number of permissions
  p = strtok_r( NULL, ",", &x );
  assert( p != NULL );
  int nr;
  scanres = sscanf (p, "%08x", &nr);
  assert (scanres == 1);
  for (int lcv = 0; lcv < nr; lcv++) {
    char name[MAX_IDENT_LENGTH+1];
    unsigned int uperm;
    p = strtok_r( NULL, ",", &x );
    assert( p != NULL );
    // %[whatev] matches members of a character class
    scanres = sscanf(p, "%"QUOTE(MAX_IDENT_LENGTH)"[^"INVALID_IDENT_CHARACTERS"]=%8x", name, &uperm);
    assert (scanres == 2);
    permissions[name] = (wfs_permission)uperm;
  }
  wfs_log(WFS_DBG_2, "permissions: %s\n",
	   format_permissions(permissions).c_str());

  close(fd);
  return true;

}

bool wfs_metadata::archive() {
  if (_filename.find('@') == std::string::npos) {
    char buf[_dir.size()+WFS_FID_NUM_HEX_BYTES+5+1+WFS_INT_NUM_DEC_BYTES];
    snprintf(buf, _dir.size()+WFS_FID_NUM_HEX_BYTES+5+1+WFS_INT_NUM_DEC_BYTES,
	     "%s@%d",_filename.c_str(),version.version_num);
    return write_helper(buf);
  } else {
    wfs_log(WFS_DBG_1, "Tried to archive metadata with an explicit version: %s.",
	     _filename.c_str());
    return false;
  }
}

bool wfs_metadata::write_to_file() {
  return write_helper(_filename.c_str());
}
bool wfs_metadata::write_helper(const char *fn) {

  int fd = creat( fn, 0666 );

  if( fd < 0 ) {
    return false;
  }

  // write the version
  char ver[WFS_INT_NUM_DEC_BYTES+1+WFS_FID_NUM_HEX_BYTES+2];
  snprintf( ver, WFS_INT_NUM_DEC_BYTES+1+WFS_FID_NUM_HEX_BYTES+2,
	    "%d,%016"PRIx64",", version.version_num, version.node_id );
  int n = write( fd, ver, strlen(ver) );
  if( n != strlen(ver) ) {
    wfs_log(WFS_DBG_1, "bad metadata (1) write to file %s\n", fn );
    close(fd);
    return false;
  }

  // permanent cues
  n = write( fd, (perm_cues + " ,").c_str(), perm_cues.length()+2 );
  if( n != perm_cues.length()+2 ) {
    wfs_log(WFS_DBG_1, "bad metadata (2) write to file %s\n", fn );
    close(fd);
    return false;
  }

  // num links
  snprintf( ver, WFS_INT_NUM_DEC_BYTES+2, "%d,", num_links );
  n = write( fd, ver, strlen(ver) );
  if( n != strlen(ver) ) {
    wfs_log(WFS_DBG_2, "bad metadata (3) write to file %s\n", fn );
    close(fd);
    return false;
  }

  /*
  if( wfs_debug(1) ) {
    printf( "Wrote out version file %08lx: %d,%016"PRIx64",%s , %08x, [reps]\n",
	    fid, version.version_num, version.node_id, perm_cues.c_str(),
	    replicas.size() );
  }
  */

  // permissions
  std::map<std::string,wfs_permission>::const_iterator i;
  char permbuf[WFS_INT_NUM_HEX_BYTES+3];
  snprintf(permbuf, WFS_INT_NUM_HEX_BYTES+2,
	   "%08x,", (unsigned int)permissions.size());
  n = write(fd, permbuf, strlen(permbuf));
  if( n != strlen(permbuf) ) {
    wfs_log(WFS_DBG_1, "bad metadata (4) write to file %s\n", fn );
    close(fd);
    return false;
  }
  for (i = permissions.begin(); i != permissions.end(); i++) {
    assert (i->first.length() <= MAX_IDENT_LENGTH);
    assert (i->first.find_first_of(INVALID_IDENT_CHARACTERS) ==
	    std::string::npos);
    n = write( fd, i->first.c_str(), i->first.length() );
    if( n != i->first.length() ) {
      wfs_log(WFS_DBG_1, "bad metadata (5) write to file %s\n", fn );
      close(fd);
      return false;
    }
    snprintf( permbuf, WFS_INT_NUM_HEX_BYTES+3, "=%08x,", i->second );
    n = write(fd, permbuf, strlen(permbuf));
    if( n != strlen(permbuf) ) {
      wfs_log(WFS_DBG_1, "bad metadata (6) write to file %s\n", fn );
      close(fd);
      return false;
    }
  }

  //  fsync(fd);
  close( fd );

  return true;
}

