/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <arpa/inet.h>
#include <config.h>
//#include <ext/hash_map>
#include <map>
#include <netdb.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>
#include <fcntl.h>
#include <dirent.h>

#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "wfs_misc.h"
#include "util/ident_util.h"
//using __gnu_cxx::hash_map;

// ~xavid: hash_map doesn't work with pthread_t keys on Mac, I think
//         because they're an opaque datatype which can't be hashed
//hash_map< pthread_t, int32_t > _wfs_thread_ids;
std::map< pthread_t, int32_t > _wfs_thread_ids;
pthread_mutex_t _wfs_id_mutex;
int32_t _wfs_thread_id_mask;

void wfs_thread_id_init( int32_t mask ) {
  pthread_mutex_init( &_wfs_id_mutex, NULL );
  _wfs_thread_id_mask = mask;
}

int32_t wfs_set_thread_id( int32_t id ) {
  pthread_mutex_lock( &_wfs_id_mutex );
  int32_t id_masked = 0;
  if( _wfs_thread_ids[pthread_self()] == 0 ) {
    if( id != 0 ) {
      id_masked = _wfs_thread_id_mask | id;
    }
    _wfs_thread_ids[pthread_self()] = id_masked;
  }
  pthread_mutex_unlock( &_wfs_id_mutex );
  return id_masked;
}

int32_t wfs_get_thread_id() {
  pthread_mutex_lock( &_wfs_id_mutex );
  int32_t id = _wfs_thread_ids[pthread_self()];
  pthread_mutex_unlock( &_wfs_id_mutex );
  return id;
}

int
TidWFSLogAugmentor::addToOutput(FILE * fp) {
  int32_t id = wfs_get_thread_id();
  if (id != 0) {
    return fprintf(fp,"<<%08x>> ", id);
  } else {
    return 0;
  }
}

float get_diff_ms( struct timeval start, struct timeval end ) {

    return
       (end.tv_sec - start.tv_sec) * 1000 +
	  (float)( (float)(end.tv_usec - start.tv_usec) / 1000);

}

#define MAXCONFBUF 0x10000

/**
 * These functions was originally copied from async/myaddrs.C in sfs, but
 * I've heavily un-dm-ified it.
 */

void cleanup_ifconf( int sock, struct ifconf ifc ) {
  free( ifc.ifc_buf );
  close( sock );
}

bool fill_ifconf( int &sock, struct ifconf &ifc ) {

  static u_int len = 512;

  if( (sock = socket( AF_INET, SOCK_DGRAM, 0 )) < 0 ) {
    wfs_log(WFS_DBG_OFF, "socket: %m\n" );
    return false;
  }

  ifc.ifc_len = MAXCONFBUF;
  ifc.ifc_buf = NULL;
  for( ;; ) {

    ifc.ifc_len = len;
    if (ifc.ifc_buf != NULL)
      free(ifc.ifc_buf);
    ifc.ifc_buf = (char *) malloc (ifc.ifc_len);
    if( ioctl( sock, SIOCGIFCONF, &ifc ) < 0 ) {

      wfs_log(WFS_DBG_OFF, "SIOCGIFCONF: %m\n" );
      cleanup_ifconf( sock, ifc );
      return false;

    }

    /* The +64 is for large addresses (e.g., IPv6), in which sa_len
     * may be greater than the struct sockaddr inside ifreq. */
    if( ifc.ifc_len + sizeof( struct ifreq ) + 64 < len ) {
      break;
    }
    len *= 2;
    if( len >= MAXCONFBUF ) {
      wfs_log(WFS_DBG_OFF, "SIOCGIFCONF: buffer too large\n" );
      cleanup_ifconf( sock, ifc );
      return false;
    }
  }

  return true;

}

bool get_next_addr( char *&curr_buf, char *end_buf, int sock,
		    in_addr &next_addr ) {

  while( curr_buf < end_buf ) {

    struct ifreq *ifrp = (struct ifreq *) curr_buf;
    struct ifreq ifr = *ifrp;

#ifndef HAVE_SA_LEN
    curr_buf += sizeof( ifr );
#else /* !HAVE_SA_LEN */
#define max(a,b) (a>b?a:b)
    curr_buf += sizeof( ifrp->ifr_name )
      + max( sizeof( ifrp->ifr_addr ), (size_t) ifrp->ifr_addr.sa_len );
#endif /* !HAVE_SA_LEN */

    if( ifrp->ifr_addr.sa_family != AF_INET ) {
      continue;
    }

    if( ioctl( sock, SIOCGIFFLAGS, &ifr) < 0 ) {
      wfs_log(WFS_DBG_OFF, "SIOCGIFFLAGS (%.*s): %m\n",
	       (int) sizeof( ifr.ifr_name ),
	       ifr.ifr_name );
      continue;
    }

    next_addr = ((struct sockaddr_in *) &ifrp->ifr_addr)->sin_addr;
    if( (ifr.ifr_flags & IFF_UP) ) {
      return true;
    }

  }

  return false;

}

const char *get_local_ip() {

  int sock;
  struct ifconf ifc;
  ifc.ifc_buf = NULL;
  if( !fill_ifconf( sock, ifc ) ) {
    return NULL;
  }

  char *curr_buf = ifc.ifc_buf;
  char *end_buf = curr_buf + ifc.ifc_len;
  in_addr next_addr;

  while( get_next_addr( curr_buf, end_buf, sock, next_addr ) ) {

    if( ntohl( next_addr.s_addr ) != INADDR_LOOPBACK ) {
      cleanup_ifconf( sock, ifc );
      return inet_ntoa( next_addr );
    }

  }

  cleanup_ifconf( sock, ifc );
  return "127.0.0.1";

}

bool is_local_ip( in_addr addr ) {

  int sock;
  struct ifconf ifc;
  ifc.ifc_buf = NULL;
  if( !fill_ifconf( sock, ifc ) ) {
    return false;
  }

  char *curr_buf = ifc.ifc_buf;
  char *end_buf = curr_buf + ifc.ifc_len;
  in_addr next_addr;

  while( get_next_addr( curr_buf, end_buf, sock, next_addr ) ) {

    if( next_addr.s_addr == addr.s_addr ) {
      cleanup_ifconf( sock, ifc );
      return true;
    }

  }

  cleanup_ifconf( sock, ifc );
  return false;

}

#if defined(__linux__) || defined(__APPLE__)
unsigned long long md5_to_long( unsigned char *md5_result )
#else
unsigned long long md5_to_long( char *md5_result )
#endif
{

  unsigned long long result;

  // copy first 32 bits to a buffer
  char int1_buf[9];
  for( int i = 0; i < 8; i++ ) {
    int1_buf[i] = md5_result[i];
  }
  int1_buf[8] = '\0';

  // copy second 32 bits to a buffer
  char int2_buf[9];
  for( int i = 8; i < 16; i++ ) {
    int2_buf[i-8] = md5_result[i];
  }
  int2_buf[8] = '\0';

  const char *res_const = int1_buf;
  unsigned int a = 0;
  int x = sscanf( res_const, "%x", &a );
  result = 0;
  result = a;
  result <<= 32;

  const char *res_const2 = int2_buf;
  unsigned int b = 0;
  x = sscanf( res_const2, "%x", &b );
  unsigned long long c = b;
  result += c;  // don't know why |= doesn't work here.  shit.

  return result;

}

void copy_file(const std::string &from, const std::string &to,
	       bool async) {
  // The simple way to copy a file; theoretically I could use configure
  // to use copyfile() if available and deal with non-POSIX, etc...
  pid_t p = fork();
  if (p == 0) {
    // Child
    execlp("cp","cp","-p",from.c_str(),to.c_str(),NULL);
    // Shouldn't get here (unless execlp fails)
    exit(errno);
  }
  // Parent
  if (!async) {
    int status;
    waitpid(p,&status,0);
    // Don't worry too much if the cp failed
    if (wfs_dbg_block(1)) {
      if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
        wfs_log(WFS_DBG_OFF, "save_old_version: cp failed with status %d!\n",
	       WEXITSTATUS(status));
      }
    }
  }
}

const int KEY_LINE_LENGTH = 381;
static char authorized_host_keys[128];
static char authorized_keys_dir[128];
static char ssh_host_rsa_key[128];
static char id_rsa_pattern[128];
static char authorized_keys_pattern[128];
char default_keydir[64];
int dummy = snprintf(default_keydir,64,
		     "%s/.wfs",getenv("HOME"));

void wfs_set_key_dir(const char *dir) {
  snprintf(authorized_host_keys,128,
	   "%s/authorized_host_keys",
	   dir);
  snprintf(ssh_host_rsa_key,128,"%s/ssh_host_rsa_key",
	   dir);
  snprintf(id_rsa_pattern,128,"%s/user_keys/%%s",
	   dir);
  snprintf(authorized_keys_dir,128,"%s/authorized_keys",
	   dir);
  snprintf(authorized_keys_pattern,128,
	   "%s/authorized_keys/%%s.pub",
	   dir);
}

void fclose_cleanup(void *arg) {
  FILE *f = (FILE *) arg;
  fclose(f);
}

void auth_free(void *arg) {
  char *a = (char *) arg;
  if( a != authorized_host_keys ) {
    delete [] a;
  }
}

bool wfs_public_key_identifies_user(const char *username,
                                    const char *public_key) {
  /**
  int oldstate;
  // disable cancellation, since it seems that pthread cleanup
  // handler might not be invoked across function pointer evaluation
  assert(pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &oldstate) == 0);
  */
  wfs_log(WFS_DBG_1, "checking whether a public key identifies user %s...\n",
	  username);
  wfs_log(WFS_DBG_2, "the public key looks like: %s\n",public_key);
  // If it's host, it needs to be a host key
  const char *authfilename;
  bool freeq = false, success = false;
  if (strcmp(username,"host") == 0) {
    authfilename = authorized_host_keys;
  } else {
    char *afn = new char[128];
    snprintf(afn, 128, authorized_keys_pattern,
	     username);
    authfilename = afn;
    freeq = true;
  }
  pthread_cleanup_push( &auth_free, (void *) authfilename );
  FILE *authfile = fopen(authfilename,"r");
  if (authfile == NULL) {
    wfs_log(WFS_DBG_1, "No authorized keys file %s!\n",authfilename);
    success = false;
    goto exit;
  }
  pthread_cleanup_push( &fclose_cleanup, (void *) authfile );
  char buf[KEY_LINE_LENGTH+MAX_IDENT_LENGTH],
    target[KEY_LINE_LENGTH];
  snprintf(target,KEY_LINE_LENGTH,
	   "ssh-rsa %s",public_key);
  while (true) {
    char *ret = fgets(buf, KEY_LINE_LENGTH+MAX_IDENT_LENGTH, authfile);

    if (ret == NULL)
      break;
    wfs_log(WFS_DBG_4, "comparing against: %s\n",buf);
    // Subtract 1 because it includes the '\0'
    if (strncmp(buf,target,KEY_LINE_LENGTH-1) == 0) {
      success = true;
      break;
    }
  }
  pthread_cleanup_pop(0);
  fclose(authfile);
  if (!success) {
    wfs_log(WFS_DBG_1, "failed identity check for %s!\n",username);
  }
 exit:
  pthread_cleanup_pop(0);
  if (freeq)
    delete [] authfilename;
  /**
  int oldstate2;
  assert(pthread_setcancelstate(oldstate, &oldstate2) == 0);
  */
  return success;
}

const char *wfs_get_private_key(const char *username) {
  if (strcmp(username,"host")==0)
    return ssh_host_rsa_key;
  else {
    static char id_rsa[128];
    snprintf(id_rsa, 128, id_rsa_pattern,username);
    return id_rsa;
  }
}

const char *wfs_get_authorized_hosts_file() {
  return authorized_host_keys;
}

const char *wfs_get_authorized_keys_dir() {
  return authorized_keys_dir;
}

void auth_update_hosts_file( const std::string hosts_file ) {


  wfs_log(WFS_DBG_1, "Updating host keys file\n" );

  std::string filename = std::string(wfs_get_authorized_hosts_file()) +
    ".tmp";
  int fd = creat( filename.c_str(), 0600 );
  if( fd >= 0 ) {
    int n = write( fd, hosts_file.c_str(), hosts_file.size() );
    close(fd);
    rename( filename.c_str(), wfs_get_authorized_hosts_file() );
  }

}

void auth_update_users_dir( const std::map< std::string, std::string > users_map ) {

  wfs_log(WFS_DBG_1, "Updating user keys directory\n" );

  std::string user_key_dir = std::string(wfs_get_authorized_keys_dir()) +
    ".tmp";

  if( mkdir( user_key_dir.c_str(), 0777 ) == 0 ) {

    std::map< std::string, std::string >::const_iterator i = users_map.begin();
    for( ; i != users_map.end(); i++ ) {

      std::string filename = user_key_dir + "/" + i->first;
      int fd = creat( filename.c_str(), 0600 );
      if( fd >= 0 ) {
	int n = write( fd, i->second.c_str(), i->second.size() );
	close(fd);
      }

    }

    // delete old directory and contents, and rename the new one
    // TODO: this is a bit dangerous . . .
    std::string command = "rm -rf \"" +
      std::string(wfs_get_authorized_keys_dir()) + "\"";
   system(command.c_str());
    rename( user_key_dir.c_str(), wfs_get_authorized_keys_dir() );

  }

}

std::string file_to_string( std::string filename ) {

  struct stat sb;
  if( stat( filename.c_str(), &sb ) != 0 ) {
    return "";
  }
  char *file_string = new char[sb.st_size];
  int fd = open(filename.c_str(), O_RDONLY);
  if( fd < 0 ) {
    delete [] file_string;
    return "";
  }

  int n = read( fd, file_string, sb.st_size );
  std::string ret = "";
  if( n >= 0 ) {
    ret = std::string( file_string, n );
  }
  delete [] file_string;
  close(fd);
  return ret;

}

std::string auth_hosts_file_string() {

  const char *authorized_hosts_file = wfs_get_authorized_hosts_file();
  return file_to_string( std::string(authorized_hosts_file) );

}

std::map<std::string, std::string> auth_user_keys_map() {

  std::string authorized_keys_dir = wfs_get_authorized_keys_dir();

  std::map<std::string, std::string> ret;

  DIR *dir = opendir( authorized_keys_dir.c_str() );
  if( dir == NULL ) {
    return ret;
  }

  struct dirent *entry;
  while( (entry = readdir(dir)) ) {

    std::string filename = authorized_keys_dir + "/" +
      std::string(entry->d_name);
    std::string file = file_to_string( filename );
    if( file != "" ) {
      ret[entry->d_name] = file;
    }

  }

  closedir(dir);
  return ret;

}
