/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang, 
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "wfs_node.h"
#include "wfs_file.h"

#include <arpa/inet.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <wfs_misc.h>

#ifdef __linux__
#include <openssl/md5.h>
#elif defined(__APPLE__)
#include <openssl/md5.h>
#else
#include <md5.h>  // requires sys/types.h
#endif /*__linux__*/

// TODO: Not protected by a lock, since I don't have a good place
// to initialize the mutex!
std::map<std::string, wfs_node_id> _wfs_node_cache;

wfs_node make_node( const char *ip, unsigned int port, std::string site_id ) {

  wfs_node new_node;
  new_node.site_id = site_id;
  new_node.addr.sin_family = AF_INET;
  new_node.addr.sin_port = htons(port);
  inet_aton(ip, &new_node.addr.sin_addr);

  // TODO: make secure?  (ip should only be so long)
  static unsigned int const addr_len = 24;
  char addr[addr_len];
  bzero( addr, addr_len );
  sprintf( addr, "%s:%d", ip, port );
  std::string s(addr);

  if( _wfs_node_cache.find(s) == _wfs_node_cache.end() ) {

    const unsigned char *addr_const = (unsigned char *) addr;
    
    MD5_CTX md5_context;
    
#if defined(__linux__) || defined(__APPLE__)
    
    MD5_Init( &md5_context );
    MD5_Update( &md5_context, addr_const, addr_len );
    unsigned char md5_result_int[WFS_CHECKSUM_LEN/2+1];
    MD5_Final( md5_result_int, &md5_context);
    
    unsigned char md5_result[WFS_CHECKSUM_LEN+1];
    // have to convert the bytes manually to hex, because there's no MD5End
    char *curr_char = (char *) md5_result;
    for( uint i = 0; i < (WFS_CHECKSUM_LEN/2); i++ ) {
      sprintf( curr_char, "%02x", md5_result_int[i] );
      curr_char += 2;
    }
    
#else
    
    MD5Init( &md5_context );
    MD5Update( &md5_context, addr_const, addr_len );
    char md5_result[WFS_CHECKSUM_LEN+1];
    MD5End( &md5_context, md5_result );
    
#endif /*__linux__*/
    _wfs_node_cache[s] = md5_to_long( md5_result );
  }
  new_node.id = _wfs_node_cache[s];

  return new_node;

}

wfs_node make_node( const char *ipandport, std::string site_id ) {
  
  char host[200];
  const char *localhost = "127.0.0.1";
  const char *port = index(ipandport, ':');
  if(port == NULL){
    memcpy(host, localhost, strlen(localhost)+1);
    port = ipandport;
  }else{
    memcpy(host, ipandport, port-ipandport);
    host[port-ipandport] = '\0';
    port++;
  }

  return make_node( host, atoi(port), site_id );

}

wfs_node make_node( host h ) {
  return make_node( inet_ntoa( h.sin().sin_addr ), htons( h.port ), h.site_id );
}
