/**
 * Copyright (c) 2009 Jeremy Stribling, Yair Sovran, Irene Zhang,
 *                    Xavid Pretzer, Jinyang Li, Frans Kaashoek, Robert Morris,
 *     		      Massachusetts Institute of Technology,
 * 		      New York University
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>

#include "wfs_misc.h"
#include "wfs_storage_mmap.h"

wfs_storage_mmap::wfs_storage_mmap( std::string dir ) : wfs_storage(dir) {
}

wfs_storage_mmap::~wfs_storage_mmap() {
}

bool wfs_storage_mmap::readfile( wfs_file_handle fh, wfs_version version,
				 unsigned long long off, int n,
				 std::string &buf ) {

  std::string f = filename( fh, version.version_num );
  int fd = open(f.c_str(), O_RDONLY);
  if( fd < 0 ) {
    return false;
  }

  // page align the offset
  unsigned long long noff = off & ~(getpagesize()-1);

  // don't read of the end of the file
  int len = n + (off-noff);
  struct stat sb;
  stat_object( fh, &sb, version.version_num );
  if( len > (int)(sb.st_size-noff) ) {
    len = (sb.st_size-noff);
  }

  if( len <= 0 ) {
    buf = "";
    close(fd);
    return true;
  }

  wfs_log(WFS_DBG_1, "About to mmap-read file %016"PRIx64" at off %lld with len %d\n",
	    fh.fid, noff, len );

  void *mapped = mmap( NULL, len, PROT_READ, MAP_SHARED, fd, noff );
  if( mapped == MAP_FAILED ) {
    perror( "Can't mmap in readfile" );
    close(fd);
    return false;
  }

  buf = std::string( (char *) mapped + (off-noff), len );

  munmap( mapped, len );
  close(fd);
  return true;

}

bool wfs_storage_mmap::writefile( wfs_file_handle fh,
				  std::vector<unsigned long long> off,
				  std::vector<std::string> buf ) {

  assert( off.size() == buf.size() );
  // find the longest length from the earliest offset
  unsigned long long min_off = 0xffffffff;
  unsigned long long max_len = 0;
  for( unsigned i = 0; i < off.size(); i++ ) {
    if( off[i] < min_off ) {
      min_off = off[i];
    }
    unsigned long long len = off[i]+buf[i].size();
    if( len > max_len ) {
      max_len = len;
    }
  }
  max_len = max_len - min_off;

  if( max_len == 0 ) {
    return true;
  }

  // can't mmap-write past the end of a file
  struct stat sb;
  stat_object( fh, &sb );
  if( max_len+min_off > (unsigned long long) sb.st_size ) {
    return wfs_storage::writefile( fh, off, buf );
  }

  std::string f = filename(fh);
  int fd = open( f.c_str(), O_RDWR );  // O_WRONLY produces an mmap perm error
  if( fd < 0 ) {
    return false;
  }

  // page align the offset
  unsigned long long nmin_off = min_off & ~(getpagesize()-1);
  max_len += (min_off-nmin_off);

  wfs_log(WFS_DBG_1, "About to mmap-write file %016"PRIx64" at off %lld with len %lld\n",
	    fh.fid, nmin_off, max_len );

  void *mapped = mmap( NULL, max_len, PROT_WRITE, MAP_SHARED, fd, nmin_off );
  if( mapped == MAP_FAILED ) {
    perror( "Can't mmap in writefile" );
    close(fd);
    return false;
  }

  for( unsigned i = 0; i < off.size(); i++ ) {
    assert( (off[i]-nmin_off)+buf[i].size() <= max_len );
    memcpy( ((char *) mapped)+off[i]-nmin_off, buf[i].data(), buf[i].size() );
  }

  munmap( mapped, max_len );

  if( fhsynclevel(fh) > 0 ) {
    fsync(fd);
  }

  close(fd);
  return true;

}

// returns true if there was a line
bool wfs_storage_mmap::find_fid( void *dirp, const std::string &name,
				 char *line, bool use_line,
				 wfs_file_id &fid, bool &higher,
				 char **line_buf, int &line_size ) {

  mmap_dir *p = (mmap_dir *) dirp;

  if( p->not_mmap != NULL ) {
    return wfs_storage::find_fid( p->not_mmap, name, line, use_line,
				  fid, higher, line_buf, line_size );
  }

  *line_buf = p->mapped + p->off;
  int first_off = p->off;

  fid = 0;
  higher = false;
  bool matching = true;
  unsigned i = 0;
  char c;
  while( p->off < p->size && (c = p->mapped[p->off++]) != '\n' ) {
    if( matching && c != name.c_str()[i] ) {
      matching = false;
      // didn't find it
      if( c > name.c_str()[i] ) {
	higher = true;
      }
      if( !use_line ) {
	do { ; } while( (c = p->mapped[p->off++]) != '\n' );
	// line not used, no need to set line_size right
	line_size = 0;
	return true;
      }
    }

    i++;

    // reach the end of the name, make sure the space matches
    if( matching && i == name.length() ) {
      c = p->mapped[p->off++];
      if( c == WFS_DIRENTRY_SEP ) {
	// found it;
	// SCNx32 is the macro for scanning uint32_t
	uint32_t fid_1;
	uint32_t fid_2;
	if( sscanf( (p->mapped+p->off), "%08"SCNx32"%08"SCNx32"\n",
		    &fid_1, &fid_2 ) == 2 ) {
	  // scan past newline
	  fid = two_ints_to_file_id( fid_1, fid_2 );
	  while( p->mapped[p->off++] != '\n' ) {}
	  line_size = p->off - first_off;
	  return true;
	}
      }
      // the entry has the name as a prefix
      matching = false;
    }

  }

  line_size = p->off - first_off;
  if( p->off >= p->size && line_size == 0 ) {
    return false;
  } else {
    return true;
  }

}

int WFS_MMAP_SIZE_LIMIT = 512*1024*1024; // 512 MB limit for mmap'd files

void * wfs_storage_mmap::open_finder( wfs_file_handle fh ) {

  // can't mmap-write past the end of a file
  struct stat sb;
  if( !stat_object( fh, &sb ) ) {
    return NULL;
  }

  if( sb.st_size == 0 || sb.st_size > WFS_MMAP_SIZE_LIMIT ) {
    mmap_dir *p = new mmap_dir();
    p->mapped = NULL;
    p->off = 0;
    p->size = 0;
    p->fd = -1;
    if( sb.st_size == 0 ) {
      p->not_mmap = NULL;
    } else {
      p->not_mmap = wfs_storage::open_finder(fh);
    }
    return p;
  }

  std::string f = filename(fh);
  int fd = open( f.c_str(), O_RDONLY );
  if( fd < 0 ) {
    return NULL;
  }

  mmap_dir *p = new mmap_dir();
  p->mapped = (char *) mmap( NULL, sb.st_size, PROT_READ, MAP_SHARED, fd, 0 );

  if( p->mapped == MAP_FAILED ) {
    delete p;
    return NULL;
  }

  p->fd = fd;
  p->off = 0;
  p->size = sb.st_size;
  p->not_mmap = NULL;

  return p;

}

void wfs_storage_mmap::close_finder( void *dirp ) {
  mmap_dir *p = (mmap_dir *) dirp;
  if( p->not_mmap != NULL ) {
    wfs_storage::close_finder( p->not_mmap );
  } else if( p->size > 0 ) {
    munmap( p->mapped, p->size );
    close(p->fd);
  }
  delete p;
}

int wfs_storage_mmap::read_from_finder( void *filep, char *buf, int size,
					const unsigned char **blockp ) {

  mmap_dir *p = (mmap_dir *) filep;

  if( p->not_mmap != NULL ) {
    return wfs_storage::read_from_finder( p->not_mmap, buf, size, blockp );
  }

  int nread = size;
  if( nread + p->off > p->size ) {
    nread = p->size - p->off;
  }
  *blockp = (unsigned char *) (p->mapped+p->off);
  p->off += nread;
  return nread;

}
